"""
Link extractor based on lxml.html
"""

from __future__ import annotations

import logging
import operator
import re
from collections.abc import Callable, Iterable
from functools import partial
from typing import TYPE_CHECKING, Any, TypeAlias, cast
from urllib.parse import urljoin, urlparse

from lxml import etree
from parsel.csstranslator import HTMLTranslator
from w3lib.html import strip_html5_whitespace
from w3lib.url import canonicalize_url, safe_url_string

from scrapy.link import Link
from scrapy.linkextractors import IGNORED_EXTENSIONS, _is_valid_url, _matches
from scrapy.utils.misc import arg_to_iter, rel_has_nofollow
from scrapy.utils.python import unique as unique_list
from scrapy.utils.response import get_base_url
from scrapy.utils.url import url_has_any_extension, url_is_from_any_domain

if TYPE_CHECKING:
    from lxml.html import HtmlElement

    from scrapy import Selector
    from scrapy.http import TextResponse


logger = logging.getLogger(__name__)

# from lxml/src/lxml/html/__init__.py
XHTML_NAMESPACE = "http://www.w3.org/1999/xhtml"

_collect_string_content = etree.XPath("string()")


def _nons(tag: Any) -> Any:
    if (
        isinstance(tag, str)
        and tag[0] == "{"
        and tag[1 : len(XHTML_NAMESPACE) + 1] == XHTML_NAMESPACE
    ):
        return tag.split("}")[-1]
    return tag


def _identity(x: Any) -> Any:
    return x


def _canonicalize_link_url(link: Link) -> str:
    return canonicalize_url(link.url, keep_fragments=True)


def _name_matches(allowed: set[str], denied: set[str], name: str) -> bool:
    """Return whether a tag or attribute *name* should be considered.

    A name matches when it is allowed and not denied. A name is allowed if it is
    listed in *allowed*, or if *allowed* contains the ``"*"`` wildcard, which
    matches every name. *denied* has precedence over *allowed*.
    """
    if name in denied:
        return False
    return "*" in allowed or name in allowed


class LxmlParserLinkExtractor:
    def __init__(
        self,
        tag: str | Callable[[str], bool] = "a",
        attr: str | Callable[[str], bool] = "href",
        process: Callable[[Any], Any] | None = None,
        unique: bool = False,
        strip: bool = True,
        canonicalized: bool = False,
    ):
        # mypy doesn't infer types for operator.* and also for partial()
        self.scan_tag: Callable[[str], bool] = (
            tag
            if callable(tag)
            else cast("Callable[[str], bool]", partial(operator.eq, tag))
        )
        self.scan_attr: Callable[[str], bool] = (
            attr
            if callable(attr)
            else cast("Callable[[str], bool]", partial(operator.eq, attr))
        )
        self.process_attr: Callable[[Any], Any] = (
            process if callable(process) else _identity
        )
        self.unique: bool = unique
        self.strip: bool = strip
        self.link_key: Callable[[Link], str] = (
            cast("Callable[[Link], str]", operator.attrgetter("url"))
            if canonicalized
            else _canonicalize_link_url
        )

    def _iter_links(
        self, document: HtmlElement
    ) -> Iterable[tuple[HtmlElement, str, str]]:
        for el in document.iter(etree.Element):
            if not self.scan_tag(_nons(el.tag)):
                continue
            attribs = el.attrib
            for attrib in attribs:
                if not self.scan_attr(attrib):
                    continue
                yield el, attrib, attribs[attrib]

    def _extract_links(
        self,
        selector: Selector,
        response_url: str,
        response_encoding: str,
        base_url: str,
    ) -> list[Link]:
        links: list[Link] = []
        # hacky way to get the underlying lxml parsed document
        for el, _, attr_val in self._iter_links(selector.root):
            # pseudo lxml.html.HtmlElement.make_links_absolute(base_url)
            try:
                if self.strip:
                    attr_val = strip_html5_whitespace(attr_val)  # noqa: PLW2901 this is intended
                attr_val = urljoin(base_url, attr_val)  # noqa: PLW2901
            except ValueError:
                continue  # skipping bogus links
            else:
                url = self.process_attr(attr_val)
                if url is None:
                    continue
            try:
                url = safe_url_string(url, encoding=response_encoding)
            except ValueError:
                logger.debug(f"Skipping extraction of link with bad URL {url!r}")
                continue

            # to fix relative links after process_value
            url = urljoin(response_url, url)
            link = Link(
                url,
                _collect_string_content(el) or "",
                nofollow=rel_has_nofollow(el.get("rel")),
            )
            links.append(link)
        return self._deduplicate_if_needed(links)

    def extract_links(self, response: TextResponse) -> list[Link]:
        base_url = get_base_url(response)
        return self._extract_links(
            response.selector, response.url, response.encoding, base_url
        )

    def _process_links(self, links: list[Link]) -> list[Link]:
        """Normalize and filter extracted links

        The subclass should override it if necessary
        """
        return self._deduplicate_if_needed(links)

    def _deduplicate_if_needed(self, links: list[Link]) -> list[Link]:
        if self.unique:
            return unique_list(links, key=self.link_key)
        return links


_Regex: TypeAlias = str | re.Pattern[str]
_RegexOrSeveral: TypeAlias = _Regex | Iterable[_Regex]


class LxmlLinkExtractor:
    r"""LxmlLinkExtractor is the recommended link extractor with handy filtering
    options. It is implemented using lxml's robust HTMLParser.

    :param allow: a single regular expression (or list of regular expressions)
        that the (absolute) urls must match in order to be extracted. If not
        given (or empty), it will match all links.
    :type allow: str or list

    :param deny: a single regular expression (or list of regular expressions)
        that the (absolute) urls must match in order to be excluded (i.e. not
        extracted). It has precedence over the ``allow`` parameter. If not
        given (or empty) it won't exclude any links.
    :type deny: str or list

    :param allow_domains: a single value or a list of string containing
        domains which will be considered for extracting the links
    :type allow_domains: str or list

    :param deny_domains: a single value or a list of strings containing
        domains which won't be considered for extracting the links
    :type deny_domains: str or list

    :param deny_extensions: a single value or list of strings containing
        extensions that should be ignored when extracting links.
        If not given, it will default to
        :data:`scrapy.linkextractors.IGNORED_EXTENSIONS`.
    :type deny_extensions: list

    :param restrict_xpaths: is an XPath (or list of XPath's) which defines
        regions inside the response where links should be extracted from.
        If given, only the text selected by those XPath will be scanned for
        links.
    :type restrict_xpaths: str or list

    :param restrict_css: a CSS selector (or list of selectors) which defines
        regions inside the response where links should be extracted from.
        Has the same behaviour as ``restrict_xpaths``.
    :type restrict_css: str or list

    :param restrict_text: a single regular expression (or list of regular
        expressions) that the link's text must match in order to be extracted.
        If not given (or empty), it will match all links. If a list of regular
        expressions is given, the link will be extracted if it matches at least
        one.
    :type restrict_text: str or list

    :param tags: a tag or a list of tags to consider when extracting links.
        Defaults to ``('a', 'area')``. Use ``'*'`` to consider every tag.
    :type tags: str or list

    :param attrs: an attribute or list of attributes which should be considered
        when looking for links to extract (only for those tags specified in the
        ``tags`` parameter). Defaults to ``('href',)``. Use ``'*'`` to consider
        every attribute.
    :type attrs: list

    :param deny_tags: a tag or a list of tags that should not be considered when
        extracting links. It has precedence over the ``tags`` parameter, so it
        can be combined with ``tags='*'`` to consider every tag except a few.
        Defaults to ``()`` (no tag is excluded).

        .. versionadded:: 2.17.0
    :type deny_tags: str or list

    :param deny_attrs: an attribute or a list of attributes that should not be
        considered when looking for links to extract. It has precedence over the
        ``attrs`` parameter, so it can be combined with ``attrs='*'`` to consider
        every attribute except a few. Defaults to ``()`` (no attribute is
        excluded).

        .. versionadded:: 2.17.0
    :type deny_attrs: str or list

    :param canonicalize: canonicalize each extracted url (using
        w3lib.url.canonicalize_url). Defaults to ``False``.
        Note that canonicalize_url is meant for duplicate checking;
        it can change the URL visible at server side, so the response can be
        different for requests with canonicalized and raw URLs. If you're
        using LinkExtractor to follow links it is more robust to
        keep the default ``canonicalize=False``.
    :type canonicalize: bool

    :param unique: whether duplicate filtering should be applied to extracted
        links.
    :type unique: bool

    :param process_value: a function which receives each value extracted from
        the tag and attributes scanned and can modify the value and return a
        new one, or return ``None`` to ignore the link altogether. If not
        given, ``process_value`` defaults to ``lambda x: x``.

        .. highlight:: html

        For example, to extract links from this code::

            <a href="javascript:goToPage('../other/page.html'); return false">Link text</a>

        .. highlight:: python

        You can use the following function in ``process_value``:

        .. code-block:: python

            def process_value(value):
                m = re.search(r"javascript:goToPage\('(.*?)'", value)
                if m:
                    return m.group(1)

        ``process_value`` is called before the filtering parameters, such as
        ``allow`` and ``deny``, which match the value that it returns. To drop
        links based on their final URL, use the ``process_links`` parameter of
        :class:`~scrapy.spiders.Rule`, which only receives links that those
        parameters kept.

    :type process_value: collections.abc.Callable

    :param strip: whether to strip whitespaces from extracted attributes.
        According to HTML5 standard, leading and trailing whitespaces
        must be stripped from ``href`` attributes of ``<a>``, ``<area>``
        and many other elements, ``src`` attribute of ``<img>``, ``<iframe>``
        elements, etc., so LinkExtractor strips space chars by default.
        Set ``strip=False`` to turn it off (e.g. if you're extracting urls
        from elements or attributes which allow leading/trailing whitespaces).
    :type strip: bool
    """

    _csstranslator = HTMLTranslator()

    def __init__(
        self,
        allow: _RegexOrSeveral = (),
        deny: _RegexOrSeveral = (),
        allow_domains: str | Iterable[str] = (),
        deny_domains: str | Iterable[str] = (),
        restrict_xpaths: str | Iterable[str] = (),
        tags: str | Iterable[str] = ("a", "area"),
        attrs: str | Iterable[str] = ("href",),
        canonicalize: bool = False,
        unique: bool = True,
        process_value: Callable[[Any], Any] | None = None,
        deny_extensions: str | Iterable[str] | None = None,
        restrict_css: str | Iterable[str] = (),
        strip: bool = True,
        restrict_text: _RegexOrSeveral | None = None,
        deny_tags: str | Iterable[str] = (),
        deny_attrs: str | Iterable[str] = (),
    ):
        tags, attrs = set(arg_to_iter(tags)), set(arg_to_iter(attrs))
        deny_tags, deny_attrs = (
            set(arg_to_iter(deny_tags)),
            set(arg_to_iter(deny_attrs)),
        )
        self.link_extractor = LxmlParserLinkExtractor(
            tag=partial(_name_matches, tags, deny_tags),
            attr=partial(_name_matches, attrs, deny_attrs),
            unique=unique,
            process=process_value,
            strip=strip,
            canonicalized=True,
        )
        self.allow_res: list[re.Pattern[str]] = self._compile_regexes(allow)
        self.deny_res: list[re.Pattern[str]] = self._compile_regexes(deny)

        self.allow_domains: set[str] = set(arg_to_iter(allow_domains))
        self.deny_domains: set[str] = set(arg_to_iter(deny_domains))

        self.restrict_xpaths: tuple[str, ...] = tuple(arg_to_iter(restrict_xpaths))
        self.restrict_xpaths += tuple(
            map(self._csstranslator.css_to_xpath, arg_to_iter(restrict_css))
        )

        if deny_extensions is None:
            deny_extensions = IGNORED_EXTENSIONS
        self.canonicalize: bool = canonicalize
        self.deny_extensions: set[str] = {"." + e for e in arg_to_iter(deny_extensions)}
        self.restrict_text: list[re.Pattern[str]] = self._compile_regexes(restrict_text)

    @staticmethod
    def _compile_regexes(value: _RegexOrSeveral | None) -> list[re.Pattern[str]]:
        return [
            x if isinstance(x, re.Pattern) else re.compile(x)
            for x in arg_to_iter(value)
        ]

    def _link_allowed(self, link: Link) -> bool:
        if not _is_valid_url(link.url):
            return False
        if self.allow_res and not _matches(link.url, self.allow_res):
            return False
        if self.deny_res and _matches(link.url, self.deny_res):
            return False
        parsed_url = urlparse(link.url)
        if self.allow_domains and not url_is_from_any_domain(
            parsed_url, self.allow_domains
        ):
            return False
        if self.deny_domains and url_is_from_any_domain(parsed_url, self.deny_domains):
            return False
        if self.deny_extensions and url_has_any_extension(
            parsed_url, self.deny_extensions
        ):
            return False
        return not self.restrict_text or _matches(link.text, self.restrict_text)

    def matches(self, url: str) -> bool:
        if self.allow_domains and not url_is_from_any_domain(url, self.allow_domains):
            return False
        if self.deny_domains and url_is_from_any_domain(url, self.deny_domains):
            return False

        allowed = (
            (regex.search(url) for regex in self.allow_res)
            if self.allow_res
            else [True]
        )
        denied = (regex.search(url) for regex in self.deny_res) if self.deny_res else ()
        return any(allowed) and not any(denied)

    def _process_links(self, links: list[Link]) -> list[Link]:
        links = [x for x in links if self._link_allowed(x)]
        if self.canonicalize:
            for link in links:
                link.url = canonicalize_url(link.url)
        return self.link_extractor._process_links(links)

    def _extract_links(self, *args: Any, **kwargs: Any) -> list[Link]:
        return self.link_extractor._extract_links(*args, **kwargs)

    def extract_links(self, response: TextResponse) -> list[Link]:
        """Returns a list of :class:`~scrapy.link.Link` objects from the
        specified :class:`response <scrapy.http.Response>`.

        Only links that match the settings passed to the ``__init__`` method of
        the link extractor are returned.

        Duplicate links are omitted if the ``unique`` attribute is set to ``True``,
        otherwise they are returned.
        """
        base_url = get_base_url(response)
        if self.restrict_xpaths:
            docs = [
                subdoc for x in self.restrict_xpaths for subdoc in response.xpath(x)
            ]
        else:
            docs = [response.selector]
        all_links = []
        for doc in docs:
            links = self._extract_links(doc, response.url, response.encoding, base_url)
            all_links.extend(self._process_links(links))
        if self.link_extractor.unique:
            return unique_list(all_links, key=self.link_extractor.link_key)
        return all_links



from __future__ import annotations

import copy
import json
import warnings
from collections.abc import Iterable, Iterator, Mapping, MutableMapping
from importlib import import_module
from logging import getLogger
from pprint import pformat
from typing import TYPE_CHECKING, Any, TypeAlias, cast

from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.settings import default_settings
from scrapy.utils.misc import load_object
from scrapy.utils.python import global_object_name

logger = getLogger(__name__)

if TYPE_CHECKING:
    from types import ModuleType

    # https://github.com/python/typing/issues/445#issuecomment-1131458824
    from _typeshed import SupportsItems

    # typing.Self requires Python 3.11
    from typing_extensions import Self

    _SettingsInput: TypeAlias = (
        SupportsItems[str, Any] | Iterable[tuple[str, Any]] | str | None
    )


SETTINGS_PRIORITIES: dict[str, int] = {
    "default": 0,
    "command": 10,
    "addon": 15,
    "project": 20,
    "spider": 30,
    "cmdline": 40,
}


def get_settings_priority(priority: int | str) -> int:
    """
    Small helper function that looks up a given string priority in the
    :attr:`~scrapy.settings.SETTINGS_PRIORITIES` dictionary and returns its
    numerical value, or directly returns a given numerical priority.
    """
    if isinstance(priority, str):
        return SETTINGS_PRIORITIES[priority]
    return priority


class SettingsAttribute:
    """Class for storing data related to settings attributes.

    This class is intended for internal usage, you should try Settings class
    for settings configuration, not this one.
    """

    def __init__(self, value: Any, priority: int):
        self.value: Any = value
        self.priority: int
        if isinstance(self.value, BaseSettings):
            self.priority = max(self.value.maxpriority(), priority)
        else:
            self.priority = priority

    def set(self, value: Any, priority: int) -> None:
        """Sets value if priority is higher or equal than current priority."""
        if priority >= self.priority:
            if isinstance(self.value, BaseSettings):
                value = BaseSettings(value, priority=priority)
            self.value = value
            self.priority = priority

    def __repr__(self) -> str:
        return f"<SettingsAttribute value={self.value!r} priority={self.priority}>"


class BaseSettings(MutableMapping[str, Any]):
    """
    Instances of this class behave like dictionaries, but store priorities
    along with their ``(key, value)`` pairs, and can be frozen (i.e. marked
    immutable).

    Key-value entries can be passed on initialization with the ``values``
    argument, and they would take the ``priority`` level (unless ``values`` is
    already an instance of :class:`~scrapy.settings.BaseSettings`, in which
    case the existing priority levels will be kept).  If the ``priority``
    argument is a string, the priority name will be looked up in
    :attr:`~scrapy.settings.SETTINGS_PRIORITIES`. Otherwise, a specific integer
    should be provided.

    Once the object is created, new settings can be loaded or updated with the
    :meth:`~scrapy.settings.BaseSettings.set` method, and can be accessed with
    the square bracket notation of dictionaries, or with the
    :meth:`~scrapy.settings.BaseSettings.get` method of the instance and its
    value conversion variants. When requesting a stored key, the value with the
    highest priority will be retrieved.
    """

    __default = object()

    def __init__(self, values: _SettingsInput = None, priority: int | str = "project"):
        self.frozen: bool = False
        self.attributes: dict[str, SettingsAttribute] = {}
        if values:
            self.update(values, priority)

    def __getitem__(self, opt_name: str) -> Any:
        if opt_name not in self:
            return None
        return self.attributes[opt_name].value

    def __contains__(self, name: Any) -> bool:
        return name in self.attributes

    def add_to_list(self, name: str, item: Any) -> None:
        """Append *item* to the :class:`list` setting with the specified *name*
        if *item* is not already in that list.

        This change is applied regardless of the priority of the *name*
        setting. The setting priority is not affected by this change either.
        """
        value: list[str] = self.getlist(name)
        if item not in value:
            self.set(name, [*value, item], self.getpriority(name) or 0)

    def remove_from_list(self, name: str, item: Any) -> None:
        """Remove *item* from the :class:`list` setting with the specified
        *name*.

        If *item* is missing, raise :exc:`ValueError`.

        This change is applied regardless of the priority of the *name*
        setting. The setting priority is not affected by this change either.
        """
        value: list[str] = self.getlist(name)
        if item not in value:
            raise ValueError(f"{item!r} not found in the {name} setting ({value!r}).")
        self.set(name, [v for v in value if v != item], self.getpriority(name) or 0)

    def get(self, name: str, default: Any = None) -> Any:  # pylint: disable=arguments-renamed
        """
        Get a setting value without affecting its original type.

        :param name: the setting name
        :type name: str

        :param default: the value to return if no setting is found
        :type default: object
        """
        if name == "CONCURRENT_REQUESTS_PER_IP" and (
            isinstance(self[name], int) and self[name] != 0
        ):
            warnings.warn(
                "The CONCURRENT_REQUESTS_PER_IP setting is deprecated, use CONCURRENT_REQUESTS_PER_DOMAIN instead.",
                ScrapyDeprecationWarning,
                stacklevel=2,
            )

        if name == "DNS_RESOLVER":
            warnings.warn(
                "The DNS_RESOLVER setting is deprecated, please use "
                "TWISTED_DNS_RESOLVER instead.",
                ScrapyDeprecationWarning,
                stacklevel=2,
            )

        return self[name] if self[name] is not None else default

    def getbool(self, name: str, default: bool = False) -> bool:
        """
        Get a setting value as a boolean.

        ``1``, ``'1'``, `True`` and ``'True'`` return ``True``,
        while ``0``, ``'0'``, ``False``, ``'False'`` and ``None`` return ``False``.

        For example, settings populated through environment variables set to
        ``'0'`` will return ``False`` when using this method.

        :param name: the setting name
        :type name: str

        :param default: the value to return if no setting is found
        :type default: object
        """
        got = self.get(name, default)
        try:
            return bool(int(got))
        except ValueError:
            if got in {"True", "true"}:
                return True
            if got in {"False", "false"}:
                return False
            raise ValueError(
                "Supported values for boolean settings "
                "are 0/1, True/False, '0'/'1', "
                "'True'/'False' and 'true'/'false'"
            ) from None

    def getint(self, name: str, default: int = 0) -> int:
        """
        Get a setting value as an int.

        :param name: the setting name
        :type name: str

        :param default: the value to return if no setting is found
        :type default: object
        """
        return int(self.get(name, default))

    def getfloat(self, name: str, default: float = 0.0) -> float:
        """
        Get a setting value as a float.

        :param name: the setting name
        :type name: str

        :param default: the value to return if no setting is found
        :type default: object
        """
        return float(self.get(name, default))

    def getlist(self, name: str, default: list[Any] | None = None) -> list[Any]:
        """
        Get a setting value as a list. If the setting original type is a list,
        a copy of it will be returned. If it's a string it will be split by
        ",". If it is an empty string, an empty list will be returned.

        For example, settings populated through environment variables set to
        ``'one,two'`` will return a list ['one', 'two'] when using this method.

        :param name: the setting name
        :type name: str

        :param default: the value to return if no setting is found
        :type default: object
        """
        value = self.get(name, default or [])
        if not value:
            return []
        if isinstance(value, str):
            value = value.split(",")
        return list(value)

    def getdict(
        self, name: str, default: dict[Any, Any] | None = None
    ) -> dict[Any, Any]:
        """
        Get a setting value as a dictionary. If the setting original type is a
        dictionary, a copy of it will be returned. If it is a string it will be
        evaluated as a JSON dictionary. In the case that it is a
        :class:`~scrapy.settings.BaseSettings` instance itself, it will be
        converted to a dictionary, containing all its current settings values
        as they would be returned by :meth:`~scrapy.settings.BaseSettings.get`,
        and losing all information about priority and mutability.

        :param name: the setting name
        :type name: str

        :param default: the value to return if no setting is found
        :type default: object
        """
        value = self.get(name, default or {})
        if isinstance(value, str):
            value = json.loads(value)
        return dict(value)

    def getdictorlist(
        self,
        name: str,
        default: dict[Any, Any] | list[Any] | tuple[Any] | None = None,
    ) -> dict[Any, Any] | list[Any]:
        """Get a setting value as either a :class:`dict` or a :class:`list`.

        If the setting is already a dict or a list, a copy of it will be
        returned.

        If it is a string it will be evaluated as JSON, or as a comma-separated
        list of strings as a fallback.

        For example, settings populated from the command line will return:

        -   ``{'key1': 'value1', 'key2': 'value2'}`` if set to
            ``'{"key1": "value1", "key2": "value2"}'``

        -   ``['one', 'two']`` if set to ``'["one", "two"]'`` or ``'one,two'``

        :param name: the setting name
        :type name: str

        :param default: the value to return if no setting is found
        :type default: object
        """
        value = self.get(name, default)
        if value is None:
            return {}
        if isinstance(value, str):
            try:
                value_loaded = json.loads(value)
                if not isinstance(value_loaded, (dict, list)):
                    raise ValueError(
                        f"JSON string for setting '{name}' must evaluate to a dict or list, "
                        f"got {type(value_loaded).__name__}: {value_loaded!r}"
                    )
                return value_loaded
            except ValueError:
                return value.split(",")
        if isinstance(value, tuple):
            return list(value)
        if not isinstance(value, (dict, list)):
            raise ValueError(
                f"Setting '{name}' must be a dict, list, tuple, or string, "
                f"got {type(value).__name__}: {value!r}"
            )
        return copy.deepcopy(value)

    def getwithbase(self, name: str) -> BaseSettings:
        """Get a composition of a dictionary-like setting and its ``_BASE``
        counterpart.

        Use
        :meth:`~scrapy.settings.BaseSettings.get_component_priority_dict_with_base`
        instead if the setting is a :ref:`component priority dictionary
        <component-priority-dictionaries>`.

        :param name: name of the dictionary-like setting
        :type name: str
        """
        if not isinstance(name, str):
            raise ValueError(f"Base setting key must be a string, got {name}")
        compbs = BaseSettings()
        compbs.update(self[name + "_BASE"])
        compbs.update(self[name])
        return compbs

    def get_component_priority_dict_with_base(self, name: str) -> BaseSettings:
        """Get a composition of a component priority dictionary setting and
        its ``_BASE`` counterpart.

        Keys are resolved to their import path for deduplication and then
        restored to their latest input representation.

        :param name: name of the component priority dictionary setting
        :type name: str
        """
        if not isinstance(name, str):
            raise ValueError(f"Base setting key must be a string, got {name}")

        normalized_keys = {}
        obj_keys = set()

        def track_loaded_key(k: Any) -> None:
            if k not in obj_keys:
                obj_keys.add(k)
                return
            logger.warning(
                f"Setting {name} contains multiple keys that refer to the "
                f"same object: {global_object_name(k)}. Only the last one will "
                f"be kept."
            )

        def normalize_key(key: Any) -> Any:
            try:
                loaded_key = load_object(key)
            except (NameError, TypeError, ValueError):
                loaded_key = key
            else:
                import_path = global_object_name(loaded_key)
                normalized_keys[import_path] = key
                key = import_path
            track_loaded_key(loaded_key)
            return key

        def restore_key(k: str) -> Any:
            return normalized_keys.get(k, k)

        result = dict(self[name + "_BASE"] or {})
        override = {normalize_key(k): v for k, v in (self[name] or {}).items()}
        result.update(override)
        return BaseSettings(
            {restore_key(k): v for k, v in result.items() if v is not None}
        )

    def getpriority(self, name: str) -> int | None:
        """
        Return the current numerical priority value of a setting, or ``None`` if
        the given ``name`` does not exist.

        :param name: the setting name
        :type name: str
        """
        if name not in self:
            return None
        return self.attributes[name].priority

    def maxpriority(self) -> int:
        """
        Return the numerical value of the highest priority present throughout
        all settings, or the numerical value for ``default`` from
        :attr:`~scrapy.settings.SETTINGS_PRIORITIES` if there are no settings
        stored.
        """
        if len(self) > 0:
            return max(cast("int", self.getpriority(name)) for name in self)
        return get_settings_priority("default")

    def replace_in_component_priority_dict(
        self,
        name: str,
        old_cls: type,
        new_cls: type,
        priority: int | None = None,
    ) -> None:
        """Replace *old_cls* with *new_cls* in the *name* :ref:`component
        priority dictionary <component-priority-dictionaries>`.

        If *old_cls* is missing, or has :data:`None` as value, :exc:`KeyError`
        is raised.

        If *old_cls* was present as an import string, even more than once,
        those keys are dropped and replaced by *new_cls*.

        If *priority* is specified, that is the value assigned to *new_cls* in
        the component priority dictionary. Otherwise, the value of *old_cls* is
        used. If *old_cls* was present multiple times (possible with import
        strings) with different values, the value assigned to *new_cls* is one
        of them, with no guarantee about which one it is.

        This change is applied regardless of the priority of the *name*
        setting. The setting priority is not affected by this change either.
        """
        component_priority_dict = self.getdict(name)
        old_priority = None
        for cls_or_path in tuple(component_priority_dict):
            if load_object(cls_or_path) != old_cls:
                continue
            if (old_priority := component_priority_dict.pop(cls_or_path)) is None:
                break
        if old_priority is None:
            raise KeyError(
                f"{old_cls} not found in the {name} setting ({component_priority_dict!r})."
            )
        component_priority_dict[new_cls] = (
            old_priority if priority is None else priority
        )
        self.set(name, component_priority_dict, priority=self.getpriority(name) or 0)

    def __setitem__(self, name: str, value: Any) -> None:
        self.set(name, value)

    def set(self, name: str, value: Any, priority: int | str = "project") -> None:
        """
        Store a key/value attribute with a given priority.

        Settings should be populated *before* the Crawler object applies them
        (in the :meth:`~scrapy.crawler.Crawler.crawl_async` or
        :meth:`~scrapy.crawler.Crawler.crawl` method), otherwise they won't
        have any effect.

        :param name: the setting name
        :type name: str

        :param value: the value to associate with the setting
        :type value: object

        :param priority: the priority of the setting. Should be a key of
            :attr:`~scrapy.settings.SETTINGS_PRIORITIES` or an integer
        :type priority: str or int
        """
        self._assert_mutability()
        priority = get_settings_priority(priority)
        if name not in self:
            if isinstance(value, SettingsAttribute):
                self.attributes[name] = value
            else:
                self.attributes[name] = SettingsAttribute(value, priority)
        else:
            self.attributes[name].set(value, priority)

    def set_in_component_priority_dict(
        self, name: str, cls: type, priority: int | None
    ) -> None:
        """Set the *cls* component in the *name* :ref:`component priority
        dictionary <component-priority-dictionaries>` setting with *priority*.

        If *cls* already exists, its value is updated.

        If *cls* was present as an import string, even more than once, those
        keys are dropped and replaced by *cls*.

        This change is applied regardless of the priority of the *name*
        setting. The setting priority is not affected by this change either.
        """
        component_priority_dict = self.getdict(name)
        for cls_or_path in tuple(component_priority_dict):
            if not isinstance(cls_or_path, str):
                continue
            _cls = load_object(cls_or_path)
            if _cls == cls:
                del component_priority_dict[cls_or_path]
        component_priority_dict[cls] = priority
        self.set(name, component_priority_dict, self.getpriority(name) or 0)

    def setdefault(  # pylint: disable=arguments-renamed
        self,
        name: str,
        default: Any = None,
        priority: int | str = "project",
    ) -> Any:
        if name not in self:
            self.set(name, default, priority)
            return default

        return self.attributes[name].value

    def setdefault_in_component_priority_dict(
        self, name: str, cls: type, priority: int | None
    ) -> None:
        """Set the *cls* component in the *name* :ref:`component priority
        dictionary <component-priority-dictionaries>` setting with *priority*
        if not already defined (even as an import string).

        If *cls* is not already defined, it is set regardless of the priority
        of the *name* setting. The setting priority is not affected by this
        change either.
        """
        component_priority_dict = self.getdict(name)
        for cls_or_path in tuple(component_priority_dict):
            if load_object(cls_or_path) == cls:
                return
        component_priority_dict[cls] = priority
        self.set(name, component_priority_dict, self.getpriority(name) or 0)

    def setdict(self, values: _SettingsInput, priority: int | str = "project") -> None:
        self.update(values, priority)

    def setmodule(
        self, module: ModuleType | str, priority: int | str = "project"
    ) -> None:
        """
        Store settings from a module with a given priority.

        This is a helper function that calls
        :meth:`~scrapy.settings.BaseSettings.set` for every globally declared
        uppercase variable of ``module`` with the provided ``priority``.

        :param module: the module or the path of the module
        :type module: types.ModuleType or str

        :param priority: the priority of the settings. Should be a key of
            :attr:`~scrapy.settings.SETTINGS_PRIORITIES` or an integer
        :type priority: str or int
        """
        self._assert_mutability()
        if isinstance(module, str):
            module = import_module(module)
        for key in dir(module):
            if key.isupper():
                self.set(key, getattr(module, key), priority)

    # BaseSettings.update() doesn't support kwargs input like MutableMapping.update().
    def update(self, values: _SettingsInput, priority: int | str = "project") -> None:  # type: ignore[override]
        """
        Store key/value pairs with a given priority.

        This is a helper function that calls
        :meth:`~scrapy.settings.BaseSettings.set` for every item of ``values``
        with the provided ``priority``.

        If ``values`` is a string, it is assumed to be JSON-encoded and parsed
        into a dict with ``json.loads()`` first. If it is a
        :class:`~scrapy.settings.BaseSettings` instance, the per-key priorities
        will be used and the ``priority`` parameter ignored. This allows
        inserting/updating settings with different priorities with a single
        command.

        :param values: the settings names and values
        :type values: dict, iterable, string or :class:`~scrapy.settings.BaseSettings`

        :param priority: the priority of the settings. Should be a key of
            :attr:`~scrapy.settings.SETTINGS_PRIORITIES` or an integer
        :type priority: str or int
        """
        self._assert_mutability()
        if isinstance(values, str):
            values = cast("dict[str, Any]", json.loads(values))
        if values is not None:
            if isinstance(values, BaseSettings):
                for name, value in values.items():
                    self.set(name, value, cast("int", values.getpriority(name)))
            else:
                items: Iterable[tuple[str, Any]]
                if hasattr(values, "items"):
                    items = cast("SupportsItems[str, Any]", values).items()
                else:
                    items = values
                for name, value in items:
                    self.set(name, value, priority)

    def delete(self, name: str, priority: int | str = "project") -> None:
        if name not in self:
            raise KeyError(name)
        self._assert_mutability()
        priority = get_settings_priority(priority)
        if priority >= cast("int", self.getpriority(name)):
            del self.attributes[name]

    def __delitem__(self, name: str) -> None:
        self._assert_mutability()
        del self.attributes[name]

    def _assert_mutability(self) -> None:
        if self.frozen:
            raise TypeError("Trying to modify an immutable Settings object")

    def copy(self) -> Self:
        """
        Make a deep copy of current settings.

        This method returns a new instance of this class,
        populated with the same values and their priorities.

        Modifications to the new object won't be reflected on the original
        settings.
        """
        return copy.deepcopy(self)

    def freeze(self) -> None:
        """
        Disable further changes to the current settings.

        After calling this method, the present state of the settings will become
        immutable. Trying to change values through the :meth:`~set` method and
        its variants won't be possible and will be alerted.
        """
        self.frozen = True

    def frozencopy(self) -> Self:
        """
        Return an immutable copy of the current settings.

        Alias for a :meth:`~freeze` call in the object returned by :meth:`copy`.
        """
        copy = self.copy()
        copy.freeze()
        return copy

    def __iter__(self) -> Iterator[str]:
        return iter(self.attributes)

    def __len__(self) -> int:
        return len(self.attributes)

    def _to_dict(self) -> dict[str, Any]:
        return {
            str(k): (v._to_dict() if isinstance(v, BaseSettings) else v)
            for k, v in self.items()
        }

    def copy_to_dict(self) -> dict[str, Any]:
        """
        Make a copy of current settings and convert to a dict.

        This method returns a new dict populated with the same values
        as the current settings.

        Modifications to the returned dict won't be reflected on the original
        settings.

        This method can be useful for example for printing settings
        in Scrapy shell.
        """
        settings = self.copy()
        return settings._to_dict()

    # https://ipython.readthedocs.io/en/stable/config/integrating.html#pretty-printing
    def _repr_pretty_(self, p: Any, cycle: bool) -> None:
        if cycle:
            p.text(repr(self))
        else:
            p.text(pformat(self.copy_to_dict()))

    def pop(self, name: str, default: Any = __default) -> Any:  # pylint: disable=arguments-renamed
        try:
            value = self.attributes[name].value
        except KeyError:
            if default is self.__default:
                raise
            return default
        del self[name]
        return value


class Settings(BaseSettings):
    """
    This object stores Scrapy settings for the configuration of internal
    components, and can be used for any further customization.

    It is a direct subclass and supports all methods of
    :class:`~scrapy.settings.BaseSettings`. Additionally, after instantiation
    of this class, the new object will have the global default settings
    described on :ref:`topics-settings-ref` already populated.
    """

    def __init__(self, values: _SettingsInput = None, priority: int | str = "project"):
        # Do not pass kwarg values here. We don't want to promote user-defined
        # dicts, and we want to update, not replace, default dicts with the
        # values given by the user
        super().__init__()
        self.setmodule(default_settings, "default")
        # Promote default dictionaries to BaseSettings instances for per-key
        # priorities
        for name, val in self.items():
            if isinstance(val, dict):
                self.set(name, BaseSettings(val, "default"), "default")
        self.update(values, priority)


def iter_default_settings() -> Iterable[tuple[str, Any]]:
    """Return the default settings as an iterator of (name, value) tuples"""
    for name in dir(default_settings):
        if name.isupper():
            yield name, getattr(default_settings, name)


def overridden_settings(
    settings: Mapping[str, Any],
) -> Iterable[tuple[str, Any]]:
    """Return an iterable of the settings that have been overridden"""
    for name, defvalue in iter_default_settings():
        value = settings[name]
        if not isinstance(defvalue, dict) and value != defvalue:
            yield name, value



from __future__ import annotations

import re
import sys
from collections.abc import AsyncGenerator, Iterable
from functools import wraps
from inspect import getmembers
from types import CoroutineType
from typing import TYPE_CHECKING, Any, ClassVar, cast
from unittest import TestCase, TestResult

from scrapy.http import Request, Response
from scrapy.utils.python import get_spec
from scrapy.utils.spider import iterate_spider_output

if TYPE_CHECKING:
    from collections.abc import Callable

    from twisted.python.failure import Failure

    from scrapy import Spider


class Contract:
    """Base class for :ref:`custom contracts <topics-contracts>`.

    *method* is the callback function to which the contract is associated.

    *args* is the list of arguments passed into the docstring, separated by
    whitespace.

    Subclasses may override :meth:`adjust_request_args`, and define a
    ``pre_process`` method or a ``post_process`` method, or both.
    """

    request_cls: type[Request] | None = None
    name: str

    def __init__(self, method: Callable[..., Any], *args: Any):
        self.testcase_pre = _create_testcase(method, f"@{self.name} pre-hook")
        self.testcase_post = _create_testcase(method, f"@{self.name} post-hook")
        self.args: tuple[Any, ...] = args

    def add_pre_hook(self, request: Request, results: TestResult) -> Request:
        if hasattr(self, "pre_process"):
            cb = request.callback
            assert cb is not None

            @wraps(cb)
            def wrapper(response: Response, **cb_kwargs: Any) -> list[Any]:
                try:
                    results.startTest(self.testcase_pre)
                    self.pre_process(response)
                    results.stopTest(self.testcase_pre)
                except AssertionError:
                    results.addFailure(self.testcase_pre, sys.exc_info())
                except Exception:
                    results.addError(self.testcase_pre, sys.exc_info())
                else:
                    results.addSuccess(self.testcase_pre)
                cb_result = cb(response, **cb_kwargs)
                if isinstance(cb_result, (AsyncGenerator, CoroutineType)):
                    if isinstance(cb_result, CoroutineType):
                        cb_result.close()
                    raise TypeError("Contracts don't support async callbacks")
                return list(cast("Iterable[Any]", iterate_spider_output(cb_result)))

            request.callback = wrapper

        return request

    def add_post_hook(self, request: Request, results: TestResult) -> Request:
        if hasattr(self, "post_process"):
            cb = request.callback
            assert cb is not None

            @wraps(cb)
            def wrapper(response: Response, **cb_kwargs: Any) -> list[Any]:
                cb_result = cb(response, **cb_kwargs)
                if isinstance(cb_result, (AsyncGenerator, CoroutineType)):
                    if isinstance(cb_result, CoroutineType):
                        cb_result.close()
                    raise TypeError("Contracts don't support async callbacks")
                output = list(cast("Iterable[Any]", iterate_spider_output(cb_result)))
                try:
                    results.startTest(self.testcase_post)
                    self.post_process(output)
                    results.stopTest(self.testcase_post)
                except AssertionError:
                    results.addFailure(self.testcase_post, sys.exc_info())
                except Exception:
                    results.addError(self.testcase_post, sys.exc_info())
                else:
                    results.addSuccess(self.testcase_post)
                return output

            request.callback = wrapper

        return request

    def adjust_request_args(self, args: dict[str, Any]) -> dict[str, Any]:
        """Receive a ``dict`` with the default arguments for the sample request
        and return it, either unmodified or with changes.

        :class:`~scrapy.Request` is used by default, but this can be changed
        with the ``request_cls`` attribute. If multiple contracts in the chain
        define this attribute, the last one is used.
        """
        return args


class ContractsManager:
    contracts: ClassVar[dict[str, type[Contract]]] = {}

    def __init__(self, contracts: Iterable[type[Contract]]):
        for contract in contracts:
            self.contracts[contract.name] = contract

    def tested_methods_from_spidercls(self, spidercls: type[Spider]) -> list[str]:
        is_method = re.compile(r"^\s*@", re.MULTILINE).search
        methods = []
        for key, value in getmembers(spidercls):
            if callable(value) and value.__doc__ and is_method(value.__doc__):
                methods.append(key)

        return methods

    def extract_contracts(self, method: Callable[..., Any]) -> list[Contract]:
        contracts: list[Contract] = []
        assert method.__doc__ is not None
        for line_ in method.__doc__.split("\n"):
            line = line_.strip()

            if line.startswith("@"):
                m = re.match(r"@(\w+)\s*(.*)", line)
                if m is None:
                    continue
                name, args = m.groups()
                args = re.split(r"\s+", args)

                contracts.append(self.contracts[name](method, *args))

        return contracts

    def from_spider(self, spider: Spider, results: TestResult) -> list[Request | None]:
        requests: list[Request | None] = []
        for method in self.tested_methods_from_spidercls(type(spider)):
            bound_method = getattr(spider, method)
            try:
                requests.append(self.from_method(bound_method, results))
            except Exception:
                case = _create_testcase(bound_method, "contract")
                results.addError(case, sys.exc_info())

        return requests

    def from_method(
        self, method: Callable[..., Any], results: TestResult
    ) -> Request | None:
        contracts = self.extract_contracts(method)
        if contracts:
            request_cls = Request
            for contract in contracts:
                if contract.request_cls is not None:
                    request_cls = contract.request_cls

            # calculate request args
            args, kwargs = get_spec(request_cls.__init__)

            # Don't filter requests to allow
            # testing different callbacks on the same URL.
            kwargs["dont_filter"] = True
            kwargs["callback"] = method

            for contract in contracts:
                kwargs = contract.adjust_request_args(kwargs)

            args.remove("self")

            # check if all positional arguments are defined in kwargs
            if set(args).issubset(set(kwargs)):
                request = request_cls(**kwargs)

                # execute pre and post hooks in order
                for contract in reversed(contracts):
                    request = contract.add_pre_hook(request, results)
                for contract in contracts:
                    request = contract.add_post_hook(request, results)

                self._clean_req(request, method, results)
                return request
        return None

    def _clean_req(
        self, request: Request, method: Callable[..., Any], results: TestResult
    ) -> None:
        """stop the request from returning objects and records any errors"""

        cb = request.callback
        assert cb is not None

        @wraps(cb)
        def cb_wrapper(response: Response, **cb_kwargs: Any) -> None:
            try:
                output = cb(response, **cb_kwargs)
                output = list(cast("Iterable[Any]", iterate_spider_output(output)))
            except Exception:
                case = _create_testcase(method, "callback")
                results.addError(case, sys.exc_info())

        def eb_wrapper(failure: Failure) -> None:
            case = _create_testcase(method, "errback")
            exc_info = failure.type, failure.value, failure.getTracebackObject()
            results.addError(case, exc_info)  # type: ignore[arg-type]

        request.callback = cb_wrapper
        request.errback = eb_wrapper


def _create_testcase(method: Callable[..., Any], desc: str) -> TestCase:
    spider = method.__self__.name  # type: ignore[attr-defined]

    class ContractTestCase(TestCase):
        def __str__(_self) -> str:  # pylint: disable=no-self-argument
            return f"[{spider}] {method.__name__} ({desc})"

    name = f"{spider}_{method.__name__}"
    setattr(ContractTestCase, name, lambda x: x)
    return ContractTestCase(name)



"""
Spider Middleware manager

See documentation in docs/topics/spider-middleware.rst
"""

from __future__ import annotations

import logging
from collections.abc import AsyncIterator, Callable, Coroutine, Iterable
from functools import wraps
from inspect import isasyncgenfunction, iscoroutine
from itertools import islice
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar
from warnings import warn

from twisted.python.failure import Failure

from scrapy import Request, Spider
from scrapy.exceptions import ScrapyDeprecationWarning, _InvalidOutput
from scrapy.http import Response
from scrapy.middleware import MiddlewareManager
from scrapy.utils.asyncgen import as_async_generator
from scrapy.utils.conf import build_component_list
from scrapy.utils.defer import (
    _process_pending_io,
    deferred_from_coro,
    maybe_deferred_to_future,
)
from scrapy.utils.python import MutableAsyncChain, global_object_name

if TYPE_CHECKING:
    from twisted.internet.defer import Deferred

    from scrapy.settings import BaseSettings


logger = logging.getLogger(__name__)


_T = TypeVar("_T")
ScrapeFunc: TypeAlias = Callable[
    [Response | Failure, Request],
    Coroutine[Any, Any, Iterable[_T] | AsyncIterator[_T]],
]


class SpiderMiddlewareManager(MiddlewareManager):
    component_name = "spider middleware"

    @classmethod
    def _get_mwlist_from_settings(cls, settings: BaseSettings) -> list[Any]:
        return build_component_list(
            settings.get_component_priority_dict_with_base("SPIDER_MIDDLEWARES")
        )

    def _add_middleware(self, mw: Any) -> None:
        if hasattr(mw, "process_spider_input"):
            self.methods["process_spider_input"].append(mw.process_spider_input)
            self._check_mw_method_spider_arg(mw.process_spider_input)

        if hasattr(mw, "process_start"):
            self.methods["process_start"].appendleft(mw.process_start)

        process_spider_output = self._get_process_spider_output(mw)
        self.methods["process_spider_output"].appendleft(process_spider_output)
        if process_spider_output is not None:
            self._check_mw_method_spider_arg(process_spider_output)

        process_spider_exception = getattr(mw, "process_spider_exception", None)
        self.methods["process_spider_exception"].appendleft(process_spider_exception)
        if process_spider_exception is not None:
            self._check_mw_method_spider_arg(process_spider_exception)

    async def _process_spider_input(
        self,
        scrape_func: ScrapeFunc[_T],
        response: Response,
        request: Request,
    ) -> Iterable[_T] | AsyncIterator[_T]:
        for method in self.methods["process_spider_input"]:
            assert method is not None
            try:
                if method in self._mw_methods_requiring_spider:
                    result = method(response=response, spider=self._spider)
                else:
                    result = method(response=response)
                if result is not None:
                    msg = (
                        f"{global_object_name(method)} must return None "
                        f"or raise an exception, got {type(result)}"
                    )
                    raise _InvalidOutput(msg)
            except _InvalidOutput:
                raise
            except Exception:
                return await scrape_func(Failure(), request)
        return await scrape_func(response, request)

    async def _evaluate_iterable(
        self,
        response: Response,
        iterable: AsyncIterator[_T],
        exception_processor_index: int,
        recover_to: MutableAsyncChain[_T],
    ) -> AsyncIterator[_T]:
        try:
            async for r in iterable:
                yield r
        except Exception as ex:
            exception_result: MutableAsyncChain[_T] = self._process_spider_exception(
                response, ex, exception_processor_index
            )
            recover_to.extend(exception_result)

    def _process_spider_exception(
        self,
        response: Response,
        exception: Exception,
        start_index: int = 0,
    ) -> MutableAsyncChain[_T]:
        # don't handle _InvalidOutput exception
        if isinstance(exception, _InvalidOutput):
            raise exception
        method_list = islice(
            self.methods["process_spider_exception"], start_index, None
        )
        for method_index, method in enumerate(method_list, start=start_index):
            if method is None:
                continue
            if method in self._mw_methods_requiring_spider:
                result = method(
                    response=response, exception=exception, spider=self._spider
                )
            else:
                result = method(response=response, exception=exception)
            if isinstance(result, (Iterable, AsyncIterator)):
                # stop exception handling by handing control over to the
                # process_spider_output chain if an iterable has been returned
                if isinstance(result, Iterable):
                    result = as_async_generator(result)
                return self._process_spider_output(response, result, method_index + 1)
            if result is None:
                continue
            msg = (
                f"{global_object_name(method)} must return None "
                f"or an iterable, got {type(result)}"
            )
            raise _InvalidOutput(msg)
        raise exception

    def _process_spider_output(
        self,
        response: Response,
        result: AsyncIterator[_T],
        start_index: int = 0,
    ) -> MutableAsyncChain[_T]:
        # items in this iterable do not need to go through the process_spider_output
        # chain, they went through it already from the process_spider_exception method
        recovered: MutableAsyncChain[_T] = MutableAsyncChain()
        method_list = islice(self.methods["process_spider_output"], start_index, None)
        for method_index, method in enumerate(method_list, start=start_index):
            if method is None:
                continue
            if method in self._mw_methods_requiring_spider:
                result = method(response=response, result=result, spider=self._spider)
            else:
                result = method(response=response, result=result)
            result = self._evaluate_iterable(
                response, result, method_index + 1, recovered
            )
        return MutableAsyncChain(result, recovered)

    async def _process_callback_output(
        self, response: Response, result: AsyncIterator[_T]
    ) -> MutableAsyncChain[_T]:
        recovered: MutableAsyncChain[_T] = MutableAsyncChain()
        result = self._evaluate_iterable(response, result, 0, recovered)
        result = self._process_spider_output(response, result)
        return MutableAsyncChain(result, recovered)

    def scrape_response(
        self,
        scrape_func: Callable[
            [Response | Failure, Request],
            Deferred[Iterable[_T] | AsyncIterator[_T]],
        ],
        response: Response,
        request: Request,
        spider: Spider,
    ) -> Deferred[MutableAsyncChain[_T]]:  # pragma: no cover
        warn(
            "SpiderMiddlewareManager.scrape_response() is deprecated, use scrape_response_async() instead",
            ScrapyDeprecationWarning,
            stacklevel=2,
        )

        @wraps(scrape_func)
        async def scrape_func_wrapped(
            response: Response | Failure, request: Request
        ) -> Iterable[_T] | AsyncIterator[_T]:
            return await maybe_deferred_to_future(scrape_func(response, request))

        self._set_compat_spider(spider)
        return deferred_from_coro(
            self.scrape_response_async(scrape_func_wrapped, response, request)
        )

    async def scrape_response_async(
        self,
        scrape_func: ScrapeFunc[_T],
        response: Response,
        request: Request,
    ) -> MutableAsyncChain[_T]:
        if not self.crawler:
            raise RuntimeError(
                "scrape_response_async() called on a SpiderMiddlewareManager"
                " instance created without a crawler."
            )
        try:
            it: Iterable[_T] | AsyncIterator[_T] = await self._process_spider_input(
                scrape_func, response, request
            )
            ait = it if isinstance(it, AsyncIterator) else as_async_generator(it)
            return await self._process_callback_output(response, ait)
        except Exception as ex:
            await _process_pending_io()
            return self._process_spider_exception(response, ex)

    async def process_start(
        self, spider: Spider | None = None
    ) -> AsyncIterator[Any] | None:
        if spider:
            if self.crawler:
                msg = (
                    "Passing a spider argument to SpiderMiddlewareManager.process_start() is deprecated"
                    " and the passed value is ignored."
                )
            else:
                msg = (
                    "Passing a spider argument to SpiderMiddlewareManager.process_start() is deprecated,"
                    " SpiderMiddlewareManager should be instantiated with a Crawler instance instead."
                )
            warn(msg, category=ScrapyDeprecationWarning, stacklevel=2)
            self._set_compat_spider(spider)
        start = self._spider.start()
        if not hasattr(start, "__aiter__"):
            if iscoroutine(start):
                start.close()
            start = self._reject_start(start)
        return await self._process_chain("process_start", start)

    async def _reject_start(self, start: Any) -> AsyncIterator[Any]:
        raise TypeError(
            f"{global_object_name(type(self._spider))}.start() must be an"
            f" asynchronous generator, i.e. an async def method with yield"
            f" statements, got {type(start)}"
        )
        yield  # pylint: disable=unreachable  # makes this method an asynchronous generator

    # This method is only needed until _async compatibility methods are removed.
    @staticmethod
    def _get_process_spider_output(mw: Any) -> Callable[..., Any] | None:
        normal_method: Callable[..., Any] | None = getattr(
            mw, "process_spider_output", None
        )
        async_method: Callable[..., Any] | None = getattr(
            mw, "process_spider_output_async", None
        )
        if not async_method:
            if normal_method and not isasyncgenfunction(normal_method):
                raise TypeError(
                    f"Middleware {global_object_name(mw.__class__)} doesn't support"
                    f" asynchronous spider output. Its process_spider_output() method"
                    f" should be an async generator function or it should additionally"
                    f" define a process_spider_output_async() method."
                )
            return normal_method
        if not normal_method:
            logger.error(
                f"Middleware {global_object_name(mw.__class__)} has"
                f" process_spider_output_async() without process_spider_output(),"
                f" skipping this method. Please rename it to process_spider_output()."
            )
            return None
        if not isasyncgenfunction(async_method):
            logger.error(
                f"{global_object_name(async_method)} is not "
                f"an async generator function, skipping this method."
            )
            return normal_method
        if isasyncgenfunction(normal_method):
            logger.error(
                f"{global_object_name(normal_method)} is an async "
                f"generator function while process_spider_output_async() exists, "
                f"skipping both methods. Please remove process_spider_output_async()."
            )
            return None
        return async_method



from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from urllib.parse import urljoin

import pytest
from twisted.internet import defer

from scrapy.core.scheduler import BaseScheduler
from scrapy.http import Request
from scrapy.spiders import Spider
from scrapy.utils.httpobj import urlparse_cached
from scrapy.utils.request import fingerprint
from scrapy.utils.test import get_crawler
from tests.utils.decorators import coroutine_test, inline_callbacks_test

if TYPE_CHECKING:
    from tests.mockserver.http import MockServer

PATHS = ["/a", "/b", "/c"]
URLS = [urljoin("https://example.org", p) for p in PATHS]


class MinimalScheduler:
    def __init__(self) -> None:
        self.requests: dict[bytes, Request] = {}

    def has_pending_requests(self) -> bool:
        return bool(self.requests)

    def enqueue_request(self, request: Request) -> bool:
        fp = fingerprint(request)
        if fp not in self.requests:
            self.requests[fp] = request
            return True
        return False

    def next_request(self) -> Request | None:
        if self.has_pending_requests():
            _, request = self.requests.popitem()
            return request
        return None


class SimpleScheduler(MinimalScheduler):
    def open(self, spider: Spider) -> defer.Deferred[str]:
        return defer.succeed("open")

    def close(self, reason: str) -> defer.Deferred[str]:
        return defer.succeed("close")

    def __len__(self) -> int:
        return len(self.requests)


class PathsSpider(Spider):
    name = "paths"

    def __init__(self, mockserver, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_urls = map(mockserver.url, PATHS)

    def parse(self, response):
        return {"path": urlparse_cached(response).path}


class InterfaceCheckMixin:
    def test_scheduler_class(self):
        assert isinstance(self.scheduler, BaseScheduler)
        assert issubclass(self.scheduler.__class__, BaseScheduler)


class TestBaseScheduler(InterfaceCheckMixin):
    def setup_method(self):
        self.scheduler = BaseScheduler()

    def test_methods(self):
        assert self.scheduler.open(Spider("foo")) is None
        assert self.scheduler.close("finished") is None
        with pytest.raises(NotImplementedError):
            self.scheduler.has_pending_requests()
        with pytest.raises(NotImplementedError):
            self.scheduler.enqueue_request(Request("https://example.org"))
        with pytest.raises(NotImplementedError):
            self.scheduler.next_request()


class TestMinimalScheduler(InterfaceCheckMixin):
    def setup_method(self):
        self.scheduler = MinimalScheduler()

    def test_open_close(self):
        with pytest.raises(AttributeError):
            self.scheduler.open(Spider("foo"))
        with pytest.raises(AttributeError):
            self.scheduler.close("finished")

    def test_len(self):
        with pytest.raises(AttributeError):
            self.scheduler.__len__()
        with pytest.raises(TypeError):
            len(self.scheduler)

    def test_enqueue_dequeue(self):
        assert not self.scheduler.has_pending_requests()
        for url in URLS:
            assert self.scheduler.enqueue_request(Request(url))
            assert not self.scheduler.enqueue_request(Request(url))
        assert self.scheduler.has_pending_requests()

        dequeued = []
        while self.scheduler.has_pending_requests():
            request = self.scheduler.next_request()
            dequeued.append(request.url)
        assert set(dequeued) == set(URLS)
        assert not self.scheduler.has_pending_requests()


class TestSimpleScheduler(InterfaceCheckMixin):
    def setup_method(self):
        self.scheduler = SimpleScheduler()

    @inline_callbacks_test
    def test_enqueue_dequeue(self):
        open_result = yield self.scheduler.open(Spider("foo"))
        assert open_result == "open"
        assert not self.scheduler.has_pending_requests()

        for url in URLS:
            assert self.scheduler.enqueue_request(Request(url))
            assert not self.scheduler.enqueue_request(Request(url))

        assert self.scheduler.has_pending_requests()
        assert len(self.scheduler) == len(URLS)

        dequeued = []
        while self.scheduler.has_pending_requests():
            request = self.scheduler.next_request()
            dequeued.append(request.url)
        assert set(dequeued) == set(URLS)

        assert not self.scheduler.has_pending_requests()
        assert len(self.scheduler) == 0

        close_result = yield self.scheduler.close("")
        assert close_result == "close"


class TestMinimalSchedulerCrawl:
    scheduler_cls = MinimalScheduler

    @coroutine_test
    async def test_crawl(
        self, caplog: pytest.LogCaptureFixture, mockserver: MockServer
    ) -> None:
        settings = {
            "SCHEDULER": self.scheduler_cls,
        }
        with caplog.at_level(logging.DEBUG):
            crawler = get_crawler(PathsSpider, settings)
            await crawler.crawl_async(mockserver)
        for path in PATHS:
            assert f"{{'path': '{path}'}}" in caplog.text
        assert f"'item_scraped_count': {len(PATHS)}" in caplog.text


class TestSimpleSchedulerCrawl(TestMinimalSchedulerCrawl):
    scheduler_cls = SimpleScheduler



from __future__ import annotations

from pathlib import Path
from time import process_time
from urllib.parse import urlparse

import pytest

from scrapy.http import HtmlResponse, Response, TextResponse
from scrapy.utils.python import to_bytes
from scrapy.utils.response import (
    get_base_url,
    get_meta_refresh,
    open_in_browser,
    response_status_message,
)


def _read_browser_output(burl: str) -> bytes:
    path = urlparse(burl).path
    if not path or not Path(path).exists():
        path = burl.replace("file://", "")
    return Path(path).read_bytes()


def test_open_in_browser():
    url = "http://www.example.com/some/page.html"
    body = (
        b"<html> <head> <title>test page</title> </head> <body>test body</body> </html>"
    )

    def browser_open(burl: str) -> bool:
        bbody = _read_browser_output(burl)
        assert b'<base href="' + to_bytes(url) + b'">' in bbody
        return True

    response = HtmlResponse(url, body=body)
    assert open_in_browser(response, _openfunc=browser_open), "Browser not called"

    resp = Response(url, body=body)
    with pytest.raises(TypeError):
        open_in_browser(resp, _openfunc=browser_open)  # type: ignore[arg-type]


def test_get_meta_refresh():
    r1 = HtmlResponse(
        "http://www.example.com",
        body=b"""
    <html>
    <head><title>Dummy</title><meta http-equiv="refresh" content="5;url=http://example.org/newpage" /></head>
    <body>blahablsdfsal&amp;</body>
    </html>""",
    )
    r2 = HtmlResponse(
        "http://www.example.com",
        body=b"""
    <html>
    <head><title>Dummy</title><noScript>
    <meta http-equiv="refresh" content="5;url=http://example.org/newpage" /></head>
    </noSCRIPT>
    <body>blahablsdfsal&amp;</body>
    </html>""",
    )
    r3 = HtmlResponse(
        "http://www.example.com",
        body=b"""
<noscript><meta http-equiv="REFRESH" content="0;url=http://www.example.com/newpage</noscript>
<script type="text/javascript">
if(!checkCookies()){
    document.write('<meta http-equiv="REFRESH" content="0;url=http://www.example.com/newpage">');
}
</script>
    """,
    )
    r4 = HtmlResponse(
        "http://www.example.com",
        body=b"""
    <html>
    <head><title>Dummy</title>
    <base href="http://www.another-domain.com/base/path/">
    <meta http-equiv="refresh" content="5;url=target.html"</head>
    <body>blahablsdfsal&amp;</body>
    </html>""",
    )
    assert get_meta_refresh(r1) == (5.0, "http://example.org/newpage")
    assert get_meta_refresh(r2) == (None, None)
    assert get_meta_refresh(r3) == (None, None)
    assert get_meta_refresh(r4) == (
        5.0,
        "http://www.another-domain.com/base/path/target.html",
    )


def test_get_base_url():
    resp = HtmlResponse(
        "http://www.example.com",
        body=b"""
    <html>
    <head><base href="http://www.example.com/img/" target="_blank"></head>
    <body>blahablsdfsal&amp;</body>
    </html>""",
    )
    assert get_base_url(resp) == "http://www.example.com/img/"

    resp2 = HtmlResponse(
        "http://www.example.com",
        body=b"""
    <html><body>blahablsdfsal&amp;</body></html>""",
    )
    assert get_base_url(resp2) == "http://www.example.com"


def test_response_status_message():
    assert response_status_message(200) == "200 OK"
    assert response_status_message(404) == "404 Not Found"
    assert response_status_message(573) == "573 Unknown Status"


@pytest.mark.parametrize(
    "body",
    [
        pytest.param(
            b"""
    <html>
        <head><title>Dummy</title></head>
        <body><p>Hello world.</p></body>
    </html>""",
            id="Simple",
        ),
        pytest.param(
            b"""
    <html>
        <head id="foo"><title>Dummy</title></head>
        <body>Hello world.</body>
    </html>""",
            id="<head> with attrs",
        ),
        pytest.param(
            b"""
    <html>
        <head><title>Dummy</title></head>
        <body>
            <header>Hello header</header>
            <p>Hello world.</p>
        </body>
    </html>""",
            id="Misleading tag",
        ),
        pytest.param(
            b"""
    <html>
        <!-- <head>Dummy comment</head> -->
        <head><title>Dummy</title></head>
        <body><p>Hello world.</p></body>
    </html>""",
            id="Misleading comment",
        ),
        pytest.param(
            b"""
    <html>
        <!--[if IE]>
        <head><title>IE head</title></head>
        <![endif]-->
        <!--[if !IE]>-->
        <head><title>Standard head</title></head>
        <!--<![endif]-->
        <body><p>Hello world.</p></body>
    </html>""",
            id="Conditional comment",
        ),
        pytest.param(
            b"""
    <html>
        <body><p>Hello world.</p></body>
    </html>""",
            id="No <head>",
        ),
        pytest.param(
            b"<p>Hello world.</p>",
            id="No <html>",
        ),
        pytest.param(
            b"""<!DOCTYPE html>
    <html>
        <head><title>Dummy</title></head>
        <body><p>Hello world.</p></body>
    </html>""",
            id="Doctype",
        ),
        pytest.param(
            b"""
    <!-- <head><base href="http://example.org"></head> -->
    <p>Hello world.</p>""",
            id="Only commented-out <head> and <base>",
        ),
    ],
)
def test_inject_base_url(body: bytes) -> None:
    url = "http://www.example.com"

    def check_base_url(burl):
        bbody = _read_browser_output(burl)
        base_tag = b'<base href="' + to_bytes(url) + b'">'
        assert bbody.count(base_tag) == 1
        index = bbody.index(base_tag)
        # The base tag is not commented out.
        assert bbody.rfind(b"<!--", 0, index) <= bbody.rfind(b"-->", 0, index)
        # The base tag comes after the doctype declaration, if any.
        assert b"<!DOCTYPE" not in bbody[index:]
        return True

    resp = HtmlResponse(url, body=body)
    assert open_in_browser(resp, _openfunc=check_base_url)


def _assert_open_in_browser_is_fast(body: bytes) -> None:
    # The exploit inputs are large enough that a vulnerable implementation
    # needs seconds to go through them, while a safe one stays in the low
    # milliseconds even on a slow interpreter.
    max_cpu_time = 0.2

    response = HtmlResponse("https://example.com", body=body)
    start_time = process_time()
    open_in_browser(response, lambda url: True)
    end_time = process_time()
    assert end_time - start_time < max_cpu_time


def test_open_in_browser_redos_comment():
    # Exploit input from
    # https://makenowjust-labs.github.io/recheck/playground/
    # for /<!--.*?-->/ (old pattern to remove comments).
    _assert_open_in_browser_is_fast(b"-><!--\x00" * 250_000 + b"->\n<!---->")


def test_open_in_browser_redos_head():
    # Exploit input from
    # https://makenowjust-labs.github.io/recheck/playground/
    # for /(<head(?:>|\s.*?>))/ (old pattern to find the head element).
    _assert_open_in_browser_is_fast(b"<head\t" * 80_000)


def test_open_in_browser_preserves_html_comments():
    url = "http://www.example.com"
    body = (
        b"<html>"
        b"<!-- preserved comment -->"
        b"<head><title>Real</title></head>"
        b"<body>content</body>"
        b"</html>"
    )

    def check(burl):
        bbody = _read_browser_output(burl)
        assert b"<!-- preserved comment -->" in bbody
        return True

    response = HtmlResponse(url, body=body)
    assert open_in_browser(response, _openfunc=check)


@pytest.mark.parametrize(
    ("base_tag", "expected_base_url"),
    [
        (b'<base href="http://real.com/">', b"http://real.com/"),
        (b'<BASE HREF="http://real.com/">', b"http://real.com/"),
        (b'<base href="/img/">', b"http://www.example.com/img/"),
        (b'<base target="_blank">', b"http://www.example.com/page.html"),
    ],
)
def test_open_in_browser_keeps_base_url_of_response(
    base_tag: bytes, expected_base_url: bytes
):
    url = "http://www.example.com/page.html"
    body = b"<html><head>" + base_tag + b"<title>T</title></head><body>hi</body></html>"

    def check(burl):
        bbody = _read_browser_output(burl)
        assert bbody.startswith(b'<base href="' + expected_base_url + b'">')
        return True

    response = HtmlResponse(url, body=body)
    assert open_in_browser(response, _openfunc=check)


def test_open_in_browser_injects_base_when_only_in_comment():
    url = "http://www.example.com"
    body = (
        b"<html>"
        b"<!-- <base href='http://other.com'> -->"
        b"<head><title>Real</title></head>"
        b"<body>content</body>"
        b"</html>"
    )

    def check(burl):
        bbody = _read_browser_output(burl)
        assert b'<base href="' + to_bytes(url) + b'">' in bbody
        return True

    response = HtmlResponse(url, body=body)
    assert open_in_browser(response, _openfunc=check)


def test_open_in_browser_injects_base_before_head_contents():
    url = "http://www.example.com"
    body = (
        b"<html>"
        b"<!--<head>comment head</head>-->"
        b"<head><title>Actual</title></head>"
        b"<body>hello</body>"
        b"</html>"
    )

    def check(burl):
        bbody = _read_browser_output(burl)
        assert bbody.count(b'<base href="' + to_bytes(url) + b'">') == 1
        base_pos = bbody.find(b'<base href="' + to_bytes(url) + b'">')
        title_pos = bbody.find(b"<title>Actual</title>")
        assert base_pos < title_pos
        return True

    response = HtmlResponse(url, body=body)
    assert open_in_browser(response, _openfunc=check)


def test_open_in_browser_text_response_uses_txt_extension():
    response = TextResponse("http://www.example.com", body=b"plain text content")

    def check(burl):
        assert burl.endswith(".txt")
        return True

    assert open_in_browser(response, _openfunc=check)


def test_open_in_browser_raises_for_unsupported_response_type():
    response = Response("http://www.example.com", body=b"binary")
    with pytest.raises(TypeError):
        open_in_browser(response, _openfunc=lambda _: True)  # type: ignore[arg-type]



"""``httpx``-based HTTP(S) download handler. Currently not recommended for production use."""

from __future__ import annotations

import ipaddress
import ssl
from contextlib import asynccontextmanager
from socket import gaierror
from typing import TYPE_CHECKING, ClassVar

from scrapy.exceptions import (
    CannotResolveHostError,
    DownloadConnectionRefusedError,
    DownloadFailedError,
    DownloadTimeoutError,
    NotConfigured,
    UnsupportedURLSchemeError,
)
from scrapy.http import Headers
from scrapy.utils._download_handlers import NullCookieJar
from scrapy.utils.python import _iter_exc_causes
from scrapy.utils.ssl import (
    _log_sslobj_debug_info,
    _make_insecure_ssl_ctx,
    _make_ssl_context,
)

from ._base_streaming import BaseStreamingDownloadHandler, _BaseResponseArgs

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from httpcore2 import AsyncNetworkStream

    from scrapy import Request
    from scrapy.crawler import Crawler


HAS_SOCKS = HAS_HTTP2 = False

try:
    try:
        import httpx2 as httpx
    except ImportError:  # pragma: no cover
        import httpx  # type: ignore[import-not-found,no-redef]
except ImportError:  # pragma: no cover
    httpx = None  # type: ignore[assignment]
else:
    # a small hack to avoid importing these optional extras unconditionally

    DOWNLOAD_FAILED_EXCEPTIONS: tuple[type[BaseException], ...] = (
        httpx.RequestError,
        httpx.InvalidURL,
    )

    try:
        import h2.exceptions

        HAS_HTTP2 = True
        DOWNLOAD_FAILED_EXCEPTIONS += (h2.exceptions.InvalidBodyLengthError,)
    except ImportError:  # pragma: no cover
        pass

    try:
        import socksio.exceptions

        HAS_SOCKS = True
        DOWNLOAD_FAILED_EXCEPTIONS += (socksio.exceptions.ProtocolError,)
    except ImportError:  # pragma: no cover
        pass


if TYPE_CHECKING:
    _Base = BaseStreamingDownloadHandler[httpx.Response]
else:
    _Base = BaseStreamingDownloadHandler


class HttpxDownloadHandler(_Base):
    experimental: ClassVar[bool] = True

    def __init__(self, crawler: Crawler):
        super().__init__(crawler)
        self._verify_certificates: bool = crawler.settings.getbool(
            "DOWNLOAD_VERIFY_CERTIFICATES"
        )
        self._enable_h2: bool = crawler.settings.getbool("HTTPX_HTTP2_ENABLED")
        if self._enable_h2 and not HAS_HTTP2:  # pragma: no cover
            raise NotConfigured(
                f"HTTP/2 support in {type(self).__name__} requires the 'httpx2[http2]' extra to be installed."
            )
        self._ssl_context: ssl.SSLContext = _make_ssl_context(crawler.settings)
        self._bind_host: str | None = self._get_bind_address_host()
        self._limits: httpx.Limits = httpx.Limits(
            # hard limit on simultaneous connections (None for no limit, which
            # is what a CONCURRENT_REQUESTS of 0 means)
            max_connections=self._pool_size_total or None,
            # total number of idle connections in the pool (extra ones are closed)
            max_keepalive_connections=self._pool_size_total or None,
        )

        self._default_client: httpx.AsyncClient = self._make_client()
        # httpx2 doesn't support per-request proxies: https://github.com/pydantic/httpx2/issues/818,
        # so we keep a pool of clients per proxy URL. LRU eviction can be added here if needed.
        self._proxy_clients: dict[str, httpx.AsyncClient] = {}

    @staticmethod
    def _check_deps_installed() -> None:
        if httpx is None:  # pragma: no cover
            raise NotConfigured(
                "HttpxDownloadHandler requires the httpx2 library to be installed."
            )

    def _make_client(self, proxy_url: str | None = None) -> httpx.AsyncClient:
        if proxy_url:
            if proxy_url.startswith("https:") and not self._verify_certificates:
                proxy_ssl_context = _make_insecure_ssl_ctx()
            else:
                proxy_ssl_context = None
            proxy = httpx.Proxy(proxy_url, ssl_context=proxy_ssl_context)
        else:
            proxy = None

        client = httpx.AsyncClient(
            cookies=NullCookieJar(),
            transport=httpx.AsyncHTTPTransport(
                verify=self._ssl_context,
                local_address=self._bind_host,
                http2=self._enable_h2,
                limits=self._limits,
                trust_env=False,
                proxy=proxy,
            ),
        )
        # https://github.com/pydantic/httpx2/issues/368
        for header_name in ("accept", "accept-encoding", "user-agent"):
            client.headers.pop(header_name, None)
        return client

    def _get_client(self, proxy_url: str | None) -> httpx.AsyncClient:
        if proxy_url is None:
            return self._default_client
        if cached := self._proxy_clients.get(proxy_url):
            return cached
        client = self._make_client(proxy_url)
        self._proxy_clients[proxy_url] = client
        return client

    @asynccontextmanager
    async def _make_request(
        self, request: Request, timeout: float
    ) -> AsyncIterator[httpx.Response]:
        proxy = self._extract_proxy_url_with_creds(request)
        if proxy and proxy.startswith("socks") and not HAS_SOCKS:  # pragma: no cover
            raise ValueError(
                f"SOCKS proxy support in {type(self).__name__} requires the 'httpx2[socks]' extra to be installed."
            )
        client = self._get_client(proxy)
        headers = self._request_headers(request).to_tuple_list()

        try:
            async with client.stream(
                request.method,
                request.url,
                content=request.body,
                headers=headers,
                timeout=timeout,
            ) as response:
                yield response
        except httpx.TimeoutException as e:
            raise DownloadTimeoutError(
                f"Getting {request.url} took longer than {timeout} seconds."
            ) from e
        except httpx.UnsupportedProtocol as e:
            raise UnsupportedURLSchemeError(str(e)) from e
        except httpx.ConnectError as e:
            if any(isinstance(c, gaierror) for c in _iter_exc_causes(e)):
                raise CannotResolveHostError(str(e)) from e
            raise DownloadConnectionRefusedError(str(e)) from e
        except httpx.ProxyError as e:
            raise DownloadConnectionRefusedError(str(e)) from e
        except DOWNLOAD_FAILED_EXCEPTIONS as e:  # pylint: disable=catching-non-exception
            raise DownloadFailedError(str(e)) from e

    @staticmethod
    def _extract_headers(response: httpx.Response) -> Headers:
        return Headers(response.headers.multi_items())

    @staticmethod
    def _build_base_response_args(
        response: httpx.Response,
        request: Request,
        headers: Headers,
    ) -> _BaseResponseArgs:
        network_stream: AsyncNetworkStream = response.extensions["network_stream"]
        server_addr = network_stream.get_extra_info("server_addr")
        ip_address = ipaddress.ip_address(server_addr[0])
        ssl_object = network_stream.get_extra_info("ssl_object")
        if isinstance(ssl_object, ssl.SSLObject):
            cert = ssl_object.getpeercert(binary_form=True)
        else:
            cert = None
        return {
            "status": response.status_code,
            "url": request.url,
            "headers": headers,
            "certificate": cert,
            "ip_address": ip_address,
            "protocol": response.http_version,
        }

    @staticmethod
    def _iter_body_chunks(response: httpx.Response) -> AsyncIterator[bytes]:
        return response.aiter_raw()

    @staticmethod
    def _is_dataloss_exception(exc: Exception) -> bool:
        return isinstance(
            exc, httpx.RemoteProtocolError
        ) and "peer closed connection without sending complete message body" in str(exc)

    def _log_tls_info(self, response: httpx.Response, request: Request) -> None:
        network_stream: AsyncNetworkStream = response.extensions["network_stream"]
        extra_ssl_object = network_stream.get_extra_info("ssl_object")
        if isinstance(extra_ssl_object, ssl.SSLObject):  # pragma: no branch
            _log_sslobj_debug_info(extra_ssl_object)

    async def close(self) -> None:
        await self._default_client.aclose()
        for client in self._proxy_clients.values():
            await client.aclose()



from __future__ import annotations

import gzip
import re
from datetime import datetime
from io import BytesIO
from logging import WARNING
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from scrapy.http import HtmlResponse, Request, Response, TextResponse, XmlResponse
from scrapy.spiders import SitemapSpider
from scrapy.utils.test import get_crawler
from tests import tests_datadir
from tests.spiders import RawResponseSpider
from tests.utils.bases.spider import TestSpiderBase
from tests.utils.crawl import crawl_items
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from tests.mockserver.http import MockServer


class RawSitemapSpider(RawResponseSpider):
    """Feeds :meth:`raw_body` to :class:`~scrapy.spiders.SitemapSpider` as a
    sitemap, so that it is fetched and followed through a regular crawl.

    Subclasses build the document in :meth:`raw_body`, typically using
    :attr:`mockserver` to point ``<loc>`` entries at real endpoints.
    """

    content_type = "application/xml"

    async def start(self):
        self.sitemap_urls = [self.raw_url]
        async for request in super().start():
            yield request


class TestSitemapSpider(TestSpiderBase):
    spider_class = SitemapSpider

    BODY = b"SITEMAP"
    f = BytesIO()
    g = gzip.GzipFile(fileobj=f, mode="w+b")
    g.write(BODY)
    g.close()
    GZBODY = f.getvalue()

    def assertSitemapBody(self, response: Response, body: bytes | None) -> None:
        crawler = get_crawler()
        spider = self.spider_class.from_crawler(crawler, "example.com")
        assert spider._get_sitemap_body(response) == body

    def test_get_sitemap_body(self):
        r: Response = XmlResponse(url="http://www.example.com/", body=self.BODY)
        self.assertSitemapBody(r, self.BODY)

        r = HtmlResponse(url="http://www.example.com/", body=self.BODY)
        self.assertSitemapBody(r, None)

        r = Response(url="http://www.example.com/favicon.ico", body=self.BODY)
        self.assertSitemapBody(r, None)

        r = XmlResponse(url="http://www.example.com/", body=b"")
        self.assertSitemapBody(r, b"")

    def test_get_sitemap_body_gzip_headers(self):
        r = Response(
            url="http://www.example.com/sitemap",
            body=self.GZBODY,
            headers={"content-type": "application/gzip"},
            request=Request("http://www.example.com/sitemap"),
        )
        self.assertSitemapBody(r, self.BODY)

    def test_get_sitemap_body_xml_url(self):
        r = TextResponse(url="http://www.example.com/sitemap.xml", body=self.BODY)
        self.assertSitemapBody(r, self.BODY)

    def test_get_sitemap_body_xml_url_compressed(self):
        r = Response(
            url="http://www.example.com/sitemap.xml.gz",
            body=self.GZBODY,
            request=Request("http://www.example.com/sitemap"),
        )
        self.assertSitemapBody(r, self.BODY)

        # .xml.gz but body decoded by HttpCompression middleware already
        r = Response(url="http://www.example.com/sitemap.xml.gz", body=self.BODY)
        self.assertSitemapBody(r, self.BODY)

    def test_get_sitemap_urls_from_robotstxt(self):
        robots = b"""# Sitemap files
Sitemap: http://example.com/sitemap.xml
Sitemap: http://example.com/sitemap-product-index.xml
Sitemap: HTTP://example.com/sitemap-uppercase.xml
Sitemap: /sitemap-relative-url.xml
"""

        r = TextResponse(url="http://www.example.com/robots.txt", body=robots)
        spider = self.spider_class("example.com")
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://example.com/sitemap.xml",
            "http://example.com/sitemap-product-index.xml",
            "http://example.com/sitemap-uppercase.xml",
            "http://www.example.com/sitemap-relative-url.xml",
        ]

    def test_get_sitemap_urls_from_robotstxt_skips_invalid_utf8_urls(self):
        robots = (
            b"User-agent: *\n"
            b"Sitemap: http://example.com/\xff.xml\n"
            b"Sitemap: http://example.com/ok.xml\n"
        )

        r = TextResponse(url="http://www.example.com/robots.txt", body=robots)
        spider = self.spider_class("example.com")

        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://example.com/ok.xml",
        ]

    def test_alternate_url_locs(self):
        sitemap = b"""<?xml version="1.0" encoding="UTF-8"?>
    <urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"
        xmlns:xhtml="http://www.w3.org/1999/xhtml">
        <url>
            <loc>http://www.example.com/english/</loc>
            <xhtml:link rel="alternate" hreflang="de"
                href="http://www.example.com/deutsch/"/>
            <xhtml:link rel="alternate" hreflang="de-ch"
                href="http://www.example.com/schweiz-deutsch/"/>
            <xhtml:link rel="alternate" hreflang="it"
                href="http://www.example.com/italiano/"/>
            <xhtml:link rel="alternate" hreflang="it"/><!-- wrong tag without href -->
        </url>
    </urlset>"""
        r = TextResponse(url="http://www.example.com/sitemap.xml", body=sitemap)
        spider = self.spider_class("example.com")
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://www.example.com/english/"
        ]

        spider.sitemap_alternate_links = True
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://www.example.com/english/",
            "http://www.example.com/deutsch/",
            "http://www.example.com/schweiz-deutsch/",
            "http://www.example.com/italiano/",
        ]

    def test_sitemap_filter(self):
        sitemap = b"""<?xml version="1.0" encoding="UTF-8"?>
    <urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"
        xmlns:xhtml="http://www.w3.org/1999/xhtml">
        <url>
            <loc>http://www.example.com/english/</loc>
            <lastmod>2010-01-01</lastmod>
        </url>
        <url>
            <loc>http://www.example.com/portuguese/</loc>
            <lastmod>2005-01-01</lastmod>
        </url>
    </urlset>"""

        class FilteredSitemapSpider(self.spider_class):  # type: ignore[name-defined,misc]
            def sitemap_filter(self, entries):
                for entry in entries:
                    date_time = datetime.strptime(entry["lastmod"], "%Y-%m-%d")
                    if date_time.year > 2008:
                        yield entry

        r = TextResponse(url="http://www.example.com/sitemap.xml", body=sitemap)
        spider = self.spider_class("example.com")
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://www.example.com/english/",
            "http://www.example.com/portuguese/",
        ]

        spider = FilteredSitemapSpider("example.com")
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://www.example.com/english/"
        ]

    def test_sitemap_filter_with_alternate_links(self):
        sitemap = b"""<?xml version="1.0" encoding="UTF-8"?>
    <urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9"
        xmlns:xhtml="http://www.w3.org/1999/xhtml">
        <url>
            <loc>http://www.example.com/english/article_1/</loc>
            <lastmod>2010-01-01</lastmod>
            <xhtml:link rel="alternate" hreflang="de"
                href="http://www.example.com/deutsch/article_1/"/>
        </url>
        <url>
            <loc>http://www.example.com/english/article_2/</loc>
            <lastmod>2015-01-01</lastmod>
        </url>
    </urlset>"""

        class FilteredSitemapSpider(self.spider_class):  # type: ignore[name-defined,misc]
            def sitemap_filter(self, entries):
                for entry in entries:
                    alternate_links = entry.get("alternate", ())
                    for link in alternate_links:
                        if "/deutsch/" in link:
                            entry["loc"] = link
                            yield entry

        r = TextResponse(url="http://www.example.com/sitemap.xml", body=sitemap)
        spider = self.spider_class("example.com")
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://www.example.com/english/article_1/",
            "http://www.example.com/english/article_2/",
        ]

        spider = FilteredSitemapSpider("example.com")
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://www.example.com/deutsch/article_1/"
        ]

    def test_sitemapindex_filter(self):
        sitemap = b"""<?xml version="1.0" encoding="UTF-8"?>
    <sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
        <sitemap>
            <loc>http://www.example.com/sitemap1.xml</loc>
            <lastmod>2004-01-01T20:00:00+00:00</lastmod>
        </sitemap>
        <sitemap>
            <loc>http://www.example.com/sitemap2.xml</loc>
            <lastmod>2005-01-01</lastmod>
        </sitemap>
    </sitemapindex>"""

        class FilteredSitemapSpider(self.spider_class):  # type: ignore[name-defined,misc]
            def sitemap_filter(self, entries):
                for entry in entries:
                    date_time = datetime.strptime(
                        entry["lastmod"].split("T")[0], "%Y-%m-%d"
                    )
                    if date_time.year > 2004:
                        yield entry

        r = TextResponse(url="http://www.example.com/sitemap.xml", body=sitemap)
        spider = self.spider_class("example.com")
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://www.example.com/sitemap1.xml",
            "http://www.example.com/sitemap2.xml",
        ]

        spider = FilteredSitemapSpider("example.com")
        assert [req.url for req in spider._parse_sitemap(r)] == [
            "http://www.example.com/sitemap2.xml"
        ]

    @pytest.mark.parametrize(
        ("rule", "result"),
        [(r"english", ["http://www.example.com/english/"]), (r"nonexistent", [])],
    )
    def test_sitemap_filter_with_rule(self, rule: str, result: list[str]):
        sitemap = b"""<?xml version="1.0" encoding="UTF-8"?>
    <urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
        <url><loc>http://www.example.com/english/</loc></url>
        <url><loc>http://www.example.com/portuguese/</loc></url>
    </urlset>"""
        r = TextResponse(url="http://www.example.com/sitemap.xml", body=sitemap)

        class _RuleSpider(self.spider_class):  # type: ignore[name-defined,misc]
            sitemap_rules = [(rule, "parse")]

        spider = _RuleSpider("example.com")
        urls = [req.url for req in spider._parse_sitemap(r)]
        assert urls == result

    @coroutine_test
    async def test_sitemap_rules_with_callable(self, mockserver: MockServer):
        # A sitemap_rules entry may hold a callable instead of a method name.
        def parse_item(response):
            yield {"url": response.url}

        class _Spider(RawSitemapSpider, self.spider_class):  # type: ignore[name-defined,misc]
            sitemap_rules = [("", parse_item)]

            def raw_body(self):
                assert self.mockserver
                loc = self.mockserver.url("/text")
                return (
                    '<?xml version="1.0" encoding="UTF-8"?>'
                    '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">'
                    f"<url><loc>{loc}</loc></url>"
                    "</urlset>"
                )

        items, _ = await crawl_items(_Spider, mockserver)
        assert items == [{"url": mockserver.url("/text")}]

    @coroutine_test
    async def test_sitemap_empty_loc(self, mockserver: MockServer):
        class _Spider(RawSitemapSpider, self.spider_class):  # type: ignore[name-defined,misc]
            def parse(self, response):
                yield {"url": response.url}

            def raw_body(self):
                assert self.mockserver
                loc = self.mockserver.url("/text")
                return (
                    '<?xml version="1.0" encoding="UTF-8"?>'
                    '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">'
                    "<url><loc></loc></url>"
                    f"<url><loc>{loc}</loc></url>"
                    "</urlset>"
                )

        items, _ = await crawl_items(_Spider, mockserver)
        assert items == [{"url": mockserver.url("/text")}]

    def test_parse_sitemap_empty_body(self, caplog: pytest.LogCaptureFixture) -> None:
        r = XmlResponse(url="http://www.example.com/sitemap.xml", body=b"")
        spider = self.spider_class("example.com")

        caplog.clear()
        with caplog.at_level(WARNING):
            results = list(spider._parse_sitemap(r))

        assert not results

        assert caplog.record_tuples == [
            (
                "scrapy.spiders.sitemap",
                WARNING,
                "Ignoring invalid sitemap: <200 http://www.example.com/sitemap.xml>",
            )
        ]

    def test_parse_sitemap_not_sitemap(self):
        body = b"""<?xml version="1.0" encoding="UTF-8"?>
    <some attr="string">
        <tag><tag3>sometext</tag3></tag>
        <tag2><tag4>sometext2</tag4></tag2>
    </some>"""
        r = XmlResponse(url="http://www.example.com/random.xml", body=body)
        spider = self.spider_class("example.com")

        results = list(spider._parse_sitemap(r))

        assert not results

    @pytest.mark.parametrize(
        ("follow", "result"),
        [
            (r"1.xml", ["http://www.example.com/sitemap1.xml"]),
            (re.compile(r"sitemap\d"), ["http://www.example.com/sitemap1.xml"]),
            (r"nonexistent", []),
        ],
    )
    def test_sitemap_follow(self, follow, result):
        sitemap = b"""<?xml version="1.0" encoding="UTF-8"?>
    <sitemapindex xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
        <sitemap>
            <loc>http://www.example.com/sitemap1.xml</loc>
        </sitemap>
    </sitemapindex>"""
        r = TextResponse(url="http://www.example.com/sitemap.xml", body=sitemap)

        class _FollowSpider(self.spider_class):  # type: ignore[name-defined,misc]
            sitemap_follow = [follow]

        spider = _FollowSpider("example.com")
        urls = [req.url for req in spider._parse_sitemap(r)]
        assert urls == result

    def test_compression_bomb_setting(self):
        settings = {"DOWNLOAD_MAXSIZE": 10_000_000}
        crawler = get_crawler(settings_dict=settings)
        spider = self.spider_class.from_crawler(crawler, "example.com")
        body_path = Path(tests_datadir, "compressed", "bomb-gzip.bin")
        body = body_path.read_bytes()
        request = Request(url="https://example.com")
        response = Response(url="https://example.com", body=body, request=request)
        assert spider._get_sitemap_body(response) is None

    @pytest.mark.filterwarnings("ignore::scrapy.exceptions.ScrapyDeprecationWarning")
    def test_compression_bomb_spider_attr(self):
        class DownloadMaxSizeSpider(self.spider_class):  # type: ignore[name-defined,misc]
            download_maxsize = 10_000_000

        crawler = get_crawler()
        spider = DownloadMaxSizeSpider.from_crawler(crawler, "example.com")
        body_path = Path(tests_datadir, "compressed", "bomb-gzip.bin")
        body = body_path.read_bytes()
        request = Request(url="https://example.com")
        response = Response(url="https://example.com", body=body, request=request)
        assert spider._get_sitemap_body(response) is None

    def test_compression_bomb_request_meta(self):
        crawler = get_crawler()
        spider = self.spider_class.from_crawler(crawler, "example.com")
        body_path = Path(tests_datadir, "compressed", "bomb-gzip.bin")
        body = body_path.read_bytes()
        request = Request(
            url="https://example.com", meta={"download_maxsize": 10_000_000}
        )
        response = Response(url="https://example.com", body=body, request=request)
        assert spider._get_sitemap_body(response) is None

    def test_download_warnsize_setting(self, caplog: pytest.LogCaptureFixture) -> None:
        settings = {"DOWNLOAD_WARNSIZE": 10_000_000}
        crawler = get_crawler(settings_dict=settings)
        spider = self.spider_class.from_crawler(crawler, "example.com")
        body_path = Path(tests_datadir, "compressed", "bomb-gzip.bin")
        body = body_path.read_bytes()
        request = Request(url="https://example.com")
        response = Response(url="https://example.com", body=body, request=request)
        caplog.clear()
        with caplog.at_level(WARNING, logger="scrapy.spiders.sitemap"):
            spider._get_sitemap_body(response)
        assert caplog.record_tuples == [
            (
                "scrapy.spiders.sitemap",
                WARNING,
                (
                    "<200 https://example.com> body size after decompression "
                    "(11511612 B) is larger than the download warning size "
                    "(10000000 B)."
                ),
            ),
        ]

    @pytest.mark.filterwarnings("ignore::scrapy.exceptions.ScrapyDeprecationWarning")
    def test_download_warnsize_spider_attr(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        class DownloadWarnSizeSpider(self.spider_class):  # type: ignore[name-defined,misc]
            download_warnsize = 10_000_000

        crawler = get_crawler()
        spider = DownloadWarnSizeSpider.from_crawler(crawler, "example.com")
        body_path = Path(tests_datadir, "compressed", "bomb-gzip.bin")
        body = body_path.read_bytes()
        request = Request(
            url="https://example.com", meta={"download_warnsize": 10_000_000}
        )
        response = Response(url="https://example.com", body=body, request=request)
        caplog.clear()
        with caplog.at_level(WARNING, logger="scrapy.spiders.sitemap"):
            spider._get_sitemap_body(response)
        assert caplog.record_tuples == [
            (
                "scrapy.spiders.sitemap",
                WARNING,
                (
                    "<200 https://example.com> body size after decompression "
                    "(11511612 B) is larger than the download warning size "
                    "(10000000 B)."
                ),
            ),
        ]

    def test_download_warnsize_request_meta(
        self, caplog: pytest.LogCaptureFixture
    ) -> None:
        crawler = get_crawler()
        spider = self.spider_class.from_crawler(crawler, "example.com")
        body_path = Path(tests_datadir, "compressed", "bomb-gzip.bin")
        body = body_path.read_bytes()
        request = Request(
            url="https://example.com", meta={"download_warnsize": 10_000_000}
        )
        response = Response(url="https://example.com", body=body, request=request)
        caplog.clear()
        with caplog.at_level(WARNING, logger="scrapy.spiders.sitemap"):
            spider._get_sitemap_body(response)
        assert caplog.record_tuples == [
            (
                "scrapy.spiders.sitemap",
                WARNING,
                (
                    "<200 https://example.com> body size after decompression "
                    "(11511612 B) is larger than the download warning size "
                    "(10000000 B)."
                ),
            ),
        ]

    @coroutine_test
    async def test_sitemap_urls(self):
        class TestSpider(self.spider_class):  # type: ignore[name-defined,misc]
            name = "test"
            sitemap_urls = ["https://toscrape.com/sitemap.xml"]

        crawler = get_crawler(TestSpider)
        spider = TestSpider.from_crawler(crawler)
        requests = [request async for request in spider.start()]

        assert len(requests) == 1
        request = requests[0]
        assert request.url == "https://toscrape.com/sitemap.xml"
        assert request.dont_filter is False
        assert request.callback == spider._parse_sitemap



"""
Base class for Scrapy spiders

See documentation in docs/topics/spiders.rst
"""

from __future__ import annotations

import logging
import warnings
from typing import TYPE_CHECKING, Any, cast

from scrapy import signals
from scrapy.exceptions import ScrapyDeprecationWarning
from scrapy.http import Request, Response
from scrapy.utils.trackref import object_ref
from scrapy.utils.url import url_is_from_spider

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from twisted.internet.defer import Deferred

    # typing.Self requires Python 3.11
    from typing_extensions import Self

    from scrapy.crawler import Crawler
    from scrapy.http.request import CallbackT
    from scrapy.settings import BaseSettings
    from scrapy.utils.log import SpiderLoggerAdapter


class Spider(object_ref):
    """Base class that any spider must subclass.

    It provides a default :meth:`start` implementation that sends
    requests based on the :attr:`start_urls` class attribute and calls the
    :meth:`parse` method for each response.
    """

    name: str
    custom_settings: dict[str, Any] | None = None

    #: Start URLs. See :meth:`start`.
    start_urls: list[str]

    def __init__(self, name: str | None = None, **kwargs: Any):
        if name is not None:
            self.name: str = name
        elif not getattr(self, "name", None):
            raise ValueError(f"{type(self).__name__} must have a name")
        self.__dict__.update(kwargs)
        if not hasattr(self, "start_urls"):
            self.start_urls: list[str] = []

    @property
    def logger(self) -> SpiderLoggerAdapter:
        # circular import
        from scrapy.utils.log import SpiderLoggerAdapter  # noqa: PLC0415

        logger = logging.getLogger(self.name)
        return SpiderLoggerAdapter(logger, {"spider": self})

    def log(self, message: Any, level: int = logging.DEBUG, **kw: Any) -> None:
        """Log the given message at the given log level

        This helper wraps a log call to the logger within the spider, but you
        can use it directly (e.g. Spider.logger.info('msg')) or use any other
        Python logger too.
        """
        warnings.warn(
            "Spider.log() is deprecated, use methods of Spider.logger instead.",
            ScrapyDeprecationWarning,
            stacklevel=2,
        )
        self.logger.log(level, message, **kw)

    @classmethod
    def from_crawler(cls, crawler: Crawler, *args: Any, **kwargs: Any) -> Self:
        spider = cls(*args, **kwargs)
        spider._set_crawler(crawler)
        return spider

    def _set_crawler(self, crawler: Crawler) -> None:
        self.crawler: Crawler = crawler
        self.settings: BaseSettings = crawler.settings
        crawler.signals.connect(self.close, signals.spider_closed)

    async def start(self) -> AsyncIterator[Any]:
        """Yield the initial :class:`~scrapy.Request` objects to send.

        .. versionadded:: 2.13

        For example:

        .. code-block:: python

            from scrapy import Request, Spider


            class MySpider(Spider):
                name = "myspider"

                async def start(self):
                    yield Request("https://toscrape.com/")

        The default implementation reads URLs from :attr:`start_urls` and
        yields a request for each with :attr:`~scrapy.Request.dont_filter`
        enabled. It is functionally equivalent to:

        .. code-block:: python

            async def start(self):
                for url in self.start_urls:
                    yield Request(url, dont_filter=True)

        You can also yield :ref:`items <topics-items>`. For example:

        .. code-block:: python

            async def start(self):
                yield {"foo": "bar"}

        To write spiders that work on Scrapy versions lower than 2.13,
        define also a synchronous ``start_requests()`` method that returns an
        iterable. For example:

        .. code-block:: python

            def start_requests(self):
                yield Request("https://toscrape.com/")

        .. seealso:: :ref:`start-requests`
        """
        for url in self.start_urls:
            yield Request(url, dont_filter=True)

    def _parse(self, response: Response, **kwargs: Any) -> Any:
        return self.parse(response, **kwargs)

    if TYPE_CHECKING:
        parse: CallbackT
    else:

        def parse(self, response: Response, **kwargs: Any) -> Any:
            """Process *response*, i.e. extract data from it and generate new
            requests.

            This is the default :ref:`callback <callbacks>`: Scrapy uses
            it for the response to any request that does not define a
            :attr:`~scrapy.Request.callback`, such as the requests that
            :meth:`start` yields by default.

            Any :attr:`~scrapy.Request.cb_kwargs` of the request are passed as
            keyword parameters.

            Spiders must define this method, unless every request that they
            send defines a callback.

            See :ref:`callback-output` about the supported return values.
            """
            raise NotImplementedError(
                f"{self.__class__.__name__}.parse callback is not defined"
            )

    @classmethod
    def update_settings(cls, settings: BaseSettings) -> None:
        settings.setdict(cls.custom_settings or {}, priority="spider")

    @classmethod
    def handles_request(cls, request: Request) -> bool:
        return url_is_from_spider(request.url, cls)

    @staticmethod
    def close(spider: Spider, reason: str) -> Deferred[None] | None:
        closed = getattr(spider, "closed", None)
        if callable(closed):
            return cast("Deferred[None] | None", closed(reason))
        return None

    def __repr__(self) -> str:
        return f"<{type(self).__name__} {self.name!r} at 0x{id(self):0x}>"


# Top-level imports
from scrapy.spiders.crawl import CrawlSpider, Rule
from scrapy.spiders.feed import CSVFeedSpider, XMLFeedSpider
from scrapy.spiders.sitemap import SitemapSpider

__all__ = [
    "CSVFeedSpider",
    "CrawlSpider",
    "Rule",
    "SitemapSpider",
    "Spider",
    "XMLFeedSpider",
]



"""Boto/botocore helpers"""

from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from scrapy.settings import BaseSettings


def is_botocore_available() -> bool:
    return find_spec("botocore") is not None


def _get_max_pool_connections(settings: BaseSettings) -> int:
    """Return the maximum number of connections that AWS clients may keep in
    their connection pool.
    """
    return settings.getint("AWS_MAX_POOL_CONNECTIONS") or settings.getint(
        "REACTOR_THREADPOOL_MAXSIZE"
    )



from __future__ import annotations

import importlib.util
import os
import signal
import sys
from io import BytesIO
from pathlib import Path
from typing import TYPE_CHECKING, Any
from unittest.mock import AsyncMock, MagicMock, patch

import pytest
from pexpect.popen_spawn import PopenSpawn

from scrapy import Spider
from scrapy.http import Request, Response
from scrapy.shell import Shell, inspect_response
from scrapy.utils.reactor import _asyncio_reactor_path
from scrapy.utils.test import get_crawler
from tests import NON_EXISTING_RESOLVABLE, tests_datadir
from tests.utils.bases.commands import TestProjectBase
from tests.utils.cmdline import proc, stop_spawn
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from scrapy.crawler import Crawler
    from tests.mockserver.http import MockServer


class TestShellCommand:
    def test_empty(self) -> None:
        _, out, _ = proc("shell", "-c", "item")
        assert "{}" in out

    def test_empty_no_reactor(self) -> None:
        _, out, _ = proc(
            "shell", "-c", "item", "--set", "TWISTED_REACTOR_ENABLED=False"
        )
        assert "{}" in out

    def test_response_body(self, mockserver: MockServer) -> None:
        _, out, _ = proc("shell", mockserver.url("/text"), "-c", "response.body")
        assert "Works" in out

    def test_response_type_text(self, mockserver: MockServer) -> None:
        _, out, _ = proc("shell", mockserver.url("/text"), "-c", "type(response)")
        assert "TextResponse" in out

    def test_response_type_html(self, mockserver: MockServer) -> None:
        _, out, _ = proc("shell", mockserver.url("/html"), "-c", "type(response)")
        assert "HtmlResponse" in out

    def test_response_selector_html(self, mockserver: MockServer) -> None:
        xpath = "response.xpath(\"//p[@class='one']/text()\").get()"
        _, out, _ = proc("shell", mockserver.url("/html"), "-c", xpath)
        assert out.strip() == "Works"

    def test_response_encoding_gb18030(self, mockserver: MockServer) -> None:
        _, out, _ = proc(
            "shell", mockserver.url("/enc-gb18030"), "-c", "response.encoding"
        )
        assert out.strip() == "gb18030"

    def test_redirect(self, mockserver: MockServer) -> None:
        _, out, _ = proc("shell", mockserver.url("/redirect"), "-c", "response.url")
        assert out.strip().endswith("/redirected")

    def test_redirect_follow_302(self, mockserver: MockServer) -> None:
        _, out, _ = proc(
            "shell",
            mockserver.url("/redirect-no-meta-refresh"),
            "-c",
            "response.status",
        )
        assert out.strip().endswith("200")

    def test_redirect_not_follow_302(self, mockserver: MockServer) -> None:
        _, out, _ = proc(
            "shell",
            "--no-redirect",
            mockserver.url("/redirect-no-meta-refresh"),
            "-c",
            "response.status",
        )
        assert out.strip().endswith("302")

    def test_fetch_redirect_follow_302(self, mockserver: MockServer) -> None:
        """Test that calling ``fetch(url)`` follows HTTP redirects by default."""
        url = mockserver.url("/redirect-no-meta-refresh")
        code = f"fetch('{url}')"
        ret, out, err = proc("shell", "-c", code)
        assert ret == 0, out
        assert "Redirecting (302)" in err
        assert "Crawled (200)" in err

    def test_fetch_redirect_not_follow_302(self, mockserver: MockServer) -> None:
        """Test that calling ``fetch(url, redirect=False)`` disables automatic redirects."""
        url = mockserver.url("/redirect-no-meta-refresh")
        code = f"fetch('{url}', redirect=False)"
        ret, out, err = proc("shell", "-c", code)
        assert ret == 0, out
        assert "Crawled (302)" in err

    def test_request_replace(self, mockserver: MockServer) -> None:
        url = mockserver.url("/text")
        code = f"fetch('{url}') or fetch(response.request.replace(method='POST'))"
        ret, out, _ = proc("shell", "-c", code)
        assert ret == 0, out

    def test_scrapy_import(self, mockserver: MockServer) -> None:
        url = mockserver.url("/text")
        code = f"fetch(scrapy.Request('{url}'))"
        ret, out, _ = proc("shell", "-c", code)
        assert ret == 0, out

    def test_local_file(self) -> None:
        filepath = Path(tests_datadir, "test_site", "index.html")
        _, out, _ = proc("shell", str(filepath), "-c", "item")
        assert "{}" in out

    def test_local_nofile(self) -> None:
        filepath = "file:///tests/sample_data/test_site/nothinghere.html"
        ret, out, err = proc("shell", filepath, "-c", "item")
        assert ret == 1, out or err
        assert "No such file or directory" in err

    def test_dns_failures(self, mockserver: MockServer) -> None:
        if NON_EXISTING_RESOLVABLE:
            pytest.skip("Non-existing hosts are resolvable")
        url = "www.somedomainthatdoesntexi.st"
        ret, out, err = proc("shell", url, "-c", "item")
        assert ret == 1, out or err
        assert "CannotResolveHostError" in err

    def test_shell_fetch_async(self, mockserver: MockServer) -> None:
        url = mockserver.url("/html")
        code = f"fetch('{url}')"
        ret, _, err = proc(
            "shell", "-c", code, "--set", f"TWISTED_REACTOR={_asyncio_reactor_path}"
        )
        assert ret == 0, err
        assert "RuntimeError: There is no current event loop in thread" not in err

    def test_shell_fetch_no_reactor(self, mockserver: MockServer) -> None:
        url = mockserver.url("/html")
        code = f"fetch('{url}')"
        ret, _, err = proc(
            "shell", "-c", code, "--set", "TWISTED_REACTOR_ENABLED=False"
        )
        assert ret == 0, err

    def test_shelp(self) -> None:
        ret, out, _ = proc("shell", "-c", "shelp()")
        assert ret == 0, out
        assert "Available Scrapy objects" in out

    def test_fetch_request_with_callbacks(self, mockserver: MockServer) -> None:
        url = mockserver.url("/text")
        code = (
            f"fetch(scrapy.Request('{url}', callback=lambda r: r, errback=lambda f: f))"
        )
        ret, out, _ = proc("shell", "-c", code)
        assert ret == 0, out


class TestShellCommandWithSpider(TestProjectBase):
    @pytest.fixture(autouse=True)
    def create_files(self, proj_path: Path) -> None:
        (proj_path / self.project_name / "spiders" / "myspider.py").write_text(
            """
import scrapy

class MySpider(scrapy.Spider):
    name = "myspider"
""",
            encoding="utf-8",
        )

    def test_spider(self, proj_path: Path, mockserver: MockServer) -> None:
        ret, out, err = proc(
            "shell",
            "--spider",
            "myspider",
            mockserver.url("/text"),
            "-c",
            "spider.name",
            cwd=proj_path,
        )
        assert ret == 0, err
        assert out.strip() == "myspider"


class TestInteractiveShell:
    # Starting an interactive shell involves an interpreter start-up, Scrapy
    # imports and shell imports, which on PyPy with coverage enabled can take
    # well over 10 seconds.
    TIMEOUT = 60

    def test_fetch(self, mockserver: MockServer) -> None:
        args = (
            sys.executable,
            "-m",
            "scrapy.cmdline",
            "shell",
        )
        env = os.environ.copy()
        env["SCRAPY_PYTHON_SHELL"] = "python"
        logfile = BytesIO()
        p = PopenSpawn(args, env=env, timeout=self.TIMEOUT)
        p.logfile_read = logfile
        p.expect_exact("Available Scrapy objects")
        p.sendline(f"fetch('{mockserver.url('/')}')")
        p.sendline("type(response)")
        p.expect_exact("HtmlResponse")
        stop_spawn(p)
        logfile.seek(0)
        assert "Traceback" not in logfile.read().decode()

    @staticmethod
    def _isolate_config(env: dict[str, str], config_home: Path) -> None:
        """Point every scrapy.cfg location (see
        :func:`scrapy.utils.conf.get_sources`) at ``config_home``.

        ``XDG_CONFIG_HOME`` is read by Scrapy on all platforms, while
        ``~/.scrapy.cfg`` goes through :func:`os.path.expanduser`, which uses
        ``HOME`` on POSIX and ``USERPROFILE`` on Windows. The working directory
        stays at the repository root (no scrapy.cfg) so subprocess coverage data
        is still collected there.
        """
        env.pop("SCRAPY_PYTHON_SHELL", None)
        env["HOME"] = str(config_home)
        env["USERPROFILE"] = str(config_home)
        env["XDG_CONFIG_HOME"] = str(config_home)

    def _run_interactive_shell(self, env: dict[str, str]) -> str:
        args = (sys.executable, "-m", "scrapy.cmdline", "shell")
        logfile = BytesIO()
        p = PopenSpawn(args, env=env, timeout=self.TIMEOUT)
        p.logfile_read = logfile
        p.expect_exact("Available Scrapy objects")
        stop_spawn(p)
        logfile.seek(0)
        return logfile.read().decode()

    @pytest.mark.skipif(
        importlib.util.find_spec("IPython") is None,
        reason="Without IPython installed, shell=python and the default both "
        "select the standard Python shell, so the setting has no observable effect.",
    )
    def test_shell_from_cfg(self, tmp_path: Path) -> None:
        config_home = tmp_path / "config"
        config_home.mkdir()
        (config_home / "scrapy.cfg").write_text("[settings]\nshell = python\n")
        env = os.environ.copy()
        self._isolate_config(env, config_home)
        args = (sys.executable, "-m", "scrapy.cmdline", "shell")
        logfile = BytesIO()
        p = PopenSpawn(args, env=env, timeout=self.TIMEOUT)
        p.logfile_read = logfile
        p.expect_exact("Available Scrapy objects")
        # The standard Python shell never imports IPython, whereas the IPython
        # shell (the default when installed) does; this confirms the configured
        # shell=python was honored, regardless of platform-specific prompts.
        p.sendline("import sys; print('IPYMODULE', 'IPython' in sys.modules)")
        p.expect_exact("IPYMODULE False")
        stop_spawn(p)
        logfile.seek(0)
        assert "Traceback" not in logfile.read().decode()

    def test_shell_default_shells(self, tmp_path: Path) -> None:
        config_home = tmp_path / "config"
        config_home.mkdir()
        env = os.environ.copy()
        self._isolate_config(env, config_home)
        assert "Traceback" not in self._run_interactive_shell(env)

    @pytest.mark.skipif(
        importlib.util.find_spec("IPython") is None, reason="IPython is not installed"
    )
    def test_shell_ipython(self, tmp_path: Path) -> None:
        # Reaching the embedded IPython shell requires selecting it explicitly,
        # since ptpython takes precedence when both are installed.
        config_home = tmp_path / "config"
        config_home.mkdir()
        env = os.environ.copy()
        self._isolate_config(env, config_home)
        env["SCRAPY_PYTHON_SHELL"] = "ipython"
        args = (sys.executable, "-m", "scrapy.cmdline", "shell")
        logfile = BytesIO()
        p = PopenSpawn(args, env=env, timeout=self.TIMEOUT)
        p.logfile_read = logfile
        p.expect_exact("Available Scrapy objects")
        p.sendline("import sys; print('IPYMODULE', 'IPython' in sys.modules)")
        p.expect_exact("IPYMODULE True")
        stop_spawn(p)
        logfile.seek(0)
        assert "Traceback" not in logfile.read().decode()


@pytest.fixture
def restore_sigint():
    """Shell.start() installs SIG_IGN as the SIGINT handler; restore it."""
    handler = signal.getsignal(signal.SIGINT)
    try:
        yield
    finally:
        signal.signal(signal.SIGINT, handler)


def _no_reactor_crawler(monkeypatch: pytest.MonkeyPatch) -> Crawler:
    """Return a crawler that reports ``TWISTED_REACTOR_ENABLED=False``.

    A genuine no-reactor crawler cannot be built while a Twisted reactor is
    installed (as it is during the test run), so we build a normal crawler and
    make its settings report the reactor as disabled, which is all the shell
    code looks at.
    """
    crawler = get_crawler()
    real_getbool = crawler.settings.getbool

    def fake_getbool(name: str, *args: Any, **kwargs: Any) -> bool:
        if name == "TWISTED_REACTOR_ENABLED":
            return False
        return real_getbool(name, *args, **kwargs)

    monkeypatch.setattr(crawler.settings, "getbool", fake_getbool)
    return crawler


@pytest.mark.requires_reactor
class TestShell:
    """Tests for :class:`~scrapy.shell.Shell` paths with no ``scrapy shell``
    command-line route: those reached through
    :func:`scrapy.shell.inspect_response` (called from spider callbacks) or only
    through direct API use, hence not covered by the subprocess tests above.
    """

    def test_populate_vars_fetch_not_available(self) -> None:
        shell = Shell(get_crawler())
        shell._inthread = False
        shell.populate_vars()
        assert "fetch" not in shell.vars

    def test_get_help_fetch_not_available(self) -> None:
        shell = Shell(get_crawler())
        shell._inthread = False
        shell.populate_vars()
        help_text = shell.get_help()
        assert "fetch(url" not in help_text
        assert "shelp()" in help_text

    def test_start_with_request(self, restore_sigint: None) -> None:
        shell = Shell(get_crawler(), code="1")
        shell.fetch = MagicMock()  # type: ignore[method-assign]
        request = Request("data:,")
        shell.start(request=request)
        shell.fetch.assert_called_once_with(request, None)

    def test_start_with_response(
        self, restore_sigint: None, capsys: pytest.CaptureFixture[str]
    ) -> None:
        shell = Shell(get_crawler(), code="response.url")
        request = Request("data:,")
        response = Response("data:,", request=request)
        shell.start(response=response)
        assert "data:," in capsys.readouterr().out
        assert shell.vars["response"] is response
        assert shell.vars["request"] is request

    @patch("scrapy.shell.start_python_console")
    def test_inspect_response(
        self, mock_console: MagicMock, restore_sigint: None
    ) -> None:
        crawler = get_crawler()
        spider = crawler._create_spider()
        response = Response("data:,", request=Request("data:,"))
        sigint_handler = signal.getsignal(signal.SIGINT)
        inspect_response(response, spider)
        mock_console.assert_called_once()
        assert signal.getsignal(signal.SIGINT) is sigint_handler

    @coroutine_test
    async def test_open_spider_explicit_spider(self) -> None:
        crawler = get_crawler()
        crawler.engine = MagicMock()
        crawler.engine.open_spider_async = AsyncMock()
        shell = Shell(crawler)
        spider = Spider.from_crawler(crawler, "test")
        await shell._open_spider(spider)
        assert shell.spider is spider
        assert crawler.spider is spider
        crawler.engine.open_spider_async.assert_called_once_with(close_if_idle=False)


@pytest.mark.only_asyncio
class TestShellNoReactor:
    @coroutine_test
    @patch("scrapy.shell.start_python_console")
    async def test_inspect_response_no_reactor(
        self,
        mock_console: MagicMock,
        restore_sigint: None,
        monkeypatch: pytest.MonkeyPatch,
    ) -> None:
        crawler = _no_reactor_crawler(monkeypatch)
        spider = crawler._create_spider()
        response = Response("data:,", request=Request("data:,"))
        inspect_response(response, spider)
        mock_console.assert_called_once()



from __future__ import annotations

from asyncio import sleep
from typing import TYPE_CHECKING, Any

import pytest

from scrapy import Spider, signals
from scrapy.utils.defer import maybe_deferred_to_future
from scrapy.utils.test import get_crawler

from .utils import twisted_sleep
from .utils.decorators import coroutine_test

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from scrapy.http import Response

ITEM_A = {"id": "a"}
ITEM_B = {"id": "b"}
ITEM_C = {"id": "c"}

SLEEP_SECONDS = 0.1


class AsyncioSleepSpiderMiddleware:
    async def process_start(self, start: AsyncIterator[Any]) -> AsyncIterator[Any]:
        await sleep(SLEEP_SECONDS)
        async for item_or_request in start:
            yield item_or_request


class NoOpSpiderMiddleware:
    async def process_start(self, start: AsyncIterator[Any]) -> AsyncIterator[Any]:
        async for item_or_request in start:
            yield item_or_request


class TwistedSleepSpiderMiddleware:
    async def process_start(self, start: AsyncIterator[Any]) -> AsyncIterator[Any]:
        await maybe_deferred_to_future(twisted_sleep(SLEEP_SECONDS))
        async for item_or_request in start:
            yield item_or_request


# Spiders and spider middlewares for TestMain._test_wrap


class ModernWrapSpider(Spider):
    name = "test"

    async def start(self) -> AsyncIterator[Any]:
        yield ITEM_B


class ModernWrapSpiderMiddleware:
    async def process_start(self, start: AsyncIterator[Any]) -> AsyncIterator[Any]:
        yield ITEM_A
        async for item_or_request in start:
            yield item_or_request
        yield ITEM_C


async def _test(
    spider_middlewares: list[type],
    spider_cls: type[Spider],
    expected_items: list[Any],
) -> None:
    actual_items = []

    def track_item(item: Any, response: Response, spider: Spider) -> None:
        actual_items.append(item)

    settings = {
        "SPIDER_MIDDLEWARES": {cls: n for n, cls in enumerate(spider_middlewares)},
    }
    crawler = get_crawler(spider_cls, settings_dict=settings)
    crawler.signals.connect(track_item, signals.item_scraped)
    await crawler.crawl_async()
    assert crawler.stats.get_value("finish_reason") == "finished"
    assert actual_items == expected_items, f"{actual_items=} != {expected_items=}"


async def _test_wrap(
    spider_middleware: type,
    spider_cls: type[Spider],
    expected_items: list[Any] | None = None,
) -> None:
    expected_items = expected_items or [ITEM_A, ITEM_B, ITEM_C]
    await _test([spider_middleware], spider_cls, expected_items)


@coroutine_test
async def test_modern_mw_modern_spider() -> None:
    await _test_wrap(ModernWrapSpiderMiddleware, ModernWrapSpider)


async def _test_sleep(spider_middlewares: list[type]) -> None:
    class TestSpider(Spider):
        name = "test"

        async def start(self) -> AsyncIterator[Any]:
            yield ITEM_A

    await _test(spider_middlewares, TestSpider, [ITEM_A])


@pytest.mark.only_asyncio
@coroutine_test
async def test_asyncio_sleep_single() -> None:
    await _test_sleep([AsyncioSleepSpiderMiddleware])


@pytest.mark.only_asyncio
@coroutine_test
async def test_asyncio_sleep_multiple() -> None:
    await _test_sleep(
        [NoOpSpiderMiddleware, AsyncioSleepSpiderMiddleware, NoOpSpiderMiddleware]
    )


@pytest.mark.requires_reactor  # needs a reactor for twisted_sleep()
@coroutine_test
async def test_twisted_sleep_single() -> None:
    await _test_sleep([TwistedSleepSpiderMiddleware])


@pytest.mark.requires_reactor  # needs a reactor for twisted_sleep()
@coroutine_test
async def test_twisted_sleep_multiple() -> None:
    await _test_sleep(
        [NoOpSpiderMiddleware, TwistedSleepSpiderMiddleware, NoOpSpiderMiddleware]
    )



from __future__ import annotations

import signal
from typing import TYPE_CHECKING, Any

import pytest

from scrapy.utils.ossignal import install_shutdown_handlers

if TYPE_CHECKING:
    from collections.abc import Generator
    from types import FrameType

SHUTDOWN_SIGNALS = [
    getattr(signal, name)
    for name in ("SIGTERM", "SIGINT", "SIGBREAK")
    if hasattr(signal, name)
]


@pytest.fixture(autouse=True)
def restore_handlers() -> Generator[None]:
    handlers = {sig: signal.getsignal(sig) for sig in SHUTDOWN_SIGNALS}
    try:
        yield
    finally:
        for sig, handler in handlers.items():
            signal.signal(sig, handler)


def shutdown(signum: int, frame: FrameType | None) -> Any:
    pass


def test_install() -> None:
    install_shutdown_handlers(shutdown)
    for sig in SHUTDOWN_SIGNALS:
        assert signal.getsignal(sig) is shutdown


def test_keeps_custom_sigint_handler() -> None:
    def custom(signum: int, frame: FrameType | None) -> Any:
        pass

    signal.signal(signal.SIGINT, custom)
    install_shutdown_handlers(shutdown, override_sigint=False)
    assert signal.getsignal(signal.SIGINT) is custom
    assert signal.getsignal(signal.SIGTERM) is shutdown


def test_overrides_default_sigint_handler() -> None:
    signal.signal(signal.SIGINT, signal.default_int_handler)
    install_shutdown_handlers(shutdown, override_sigint=False)
    assert signal.getsignal(signal.SIGINT) is shutdown



import sys

from OpenSSL import __version__ as PYOPENSSL_VERSION_STRING
from packaging.version import Version
from parsel import __version__ as PARSEL_VERSION_STRING
from twisted import version as TWISTED_VERSION
from twisted.python.versions import Version as TxVersion

# improved urllib.robotparser, https://github.com/python/cpython/pull/149374
STDLIB_IMPROVED_ROBOTFILEPARSER = sys.version_info >= (3, 14, 5) or (
    (3, 13, 14) <= sys.version_info < (3, 14)
)

TWISTED_FAILURE_HAS_STACK = TWISTED_VERSION < TxVersion("twisted", 24, 10, 0)
# changes to private _sslverify code, https://github.com/twisted/twisted/pull/12506
TWISTED_TLS_NEW_IMPL = TWISTED_VERSION >= TxVersion("twisted", 26, 4, 0)
# lowerMaximumSecurityTo off-by-1, https://github.com/twisted/twisted/issues/10232
TWISTED_TLS_LIMITS_OFFBY1 = TWISTED_VERSION < TxVersion("twisted", 26, 4, 0)

PARSEL_VERSION = Version(PARSEL_VERSION_STRING)
# Selector.jmespath() support
PARSEL_SUPPORTS_JMESPATH = PARSEL_VERSION >= Version("1.8.0")

PYOPENSSL_VERSION = Version(PYOPENSSL_VERSION_STRING)
# SSL.Context.set_cipher_list() creates a temporary connection, making the context immutable
PYOPENSSL_SET_CIPHER_LIST_TMP_CONN = PYOPENSSL_VERSION < Version("25.2.0")



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from scrapy import signals
from scrapy.utils.test import get_crawler

if TYPE_CHECKING:
    from scrapy.crawler import Crawler
    from scrapy.spiders import Spider
    from tests.mockserver.http import MockServer


async def crawl_items(
    spider_cls: type[Spider], mockserver: MockServer, **kwargs: Any
) -> tuple[list[Any], Crawler]:
    """Run *spider_cls* against *mockserver* and return the scraped items along
    with the crawler, which gives tests access to the resulting stats."""
    items: list[Any] = []

    def collect(item: Any) -> None:
        items.append(item)

    crawler = get_crawler(spider_cls)
    crawler.signals.connect(collect, signals.item_scraped)
    await crawler.crawl_async(mockserver=mockserver, **kwargs)
    return items, crawler



from __future__ import annotations

import warnings
from functools import partial
from typing import TYPE_CHECKING, Any
from unittest import mock

import pytest

from scrapy.utils.misc import (
    is_generator_with_return_value,
    warn_on_generator_with_return_value,
)

if TYPE_CHECKING:
    from collections.abc import Generator

    from scrapy import Spider


def _indentation_error(*args, **kwargs):
    raise IndentationError


def top_level_return_something():
    """
    docstring
    """
    url = """
https://example.org
"""
    yield url
    return 1


def top_level_return_none():
    """
    docstring
    """
    url = """
https://example.org
"""
    yield url


def generator_that_returns_stuff() -> Generator[int, None, int]:
    yield 1
    yield 2
    return 3


@pytest.fixture
def mock_spider() -> Spider:
    class MockSettings:
        def __init__(self, settings_dict: dict[str, Any] | None = None):
            self.settings_dict = settings_dict or {
                "WARN_ON_GENERATOR_RETURN_VALUE": True
            }

        def getbool(self, name, default=False):
            return self.settings_dict.get(name, default)

    class MockSpider:
        def __init__(self) -> None:
            self.settings = MockSettings()

    return MockSpider()  # type: ignore[return-value]


def test_generators_return_something(mock_spider):
    def f1():
        yield 1
        return 2

    def g1():
        yield 1
        return "asdf"

    def h1():
        yield 1

        def helper() -> int:
            return 0

        yield helper()
        return 2

    def i1():
        """
        docstring
        """
        url = """
https://example.org
    """
        yield url
        return 1

    assert is_generator_with_return_value(top_level_return_something)
    assert is_generator_with_return_value(f1)
    assert is_generator_with_return_value(g1)
    assert is_generator_with_return_value(h1)
    assert is_generator_with_return_value(i1)

    with pytest.warns(
        UserWarning,
        match='The "MockSpider.top_level_return_something" method is a generator',
    ):
        warn_on_generator_with_return_value(mock_spider, top_level_return_something)
    with pytest.warns(UserWarning, match='The "MockSpider.f1" method is a generator'):
        warn_on_generator_with_return_value(mock_spider, f1)
    with pytest.warns(UserWarning, match='The "MockSpider.g1" method is a generator'):
        warn_on_generator_with_return_value(mock_spider, g1)
    with pytest.warns(UserWarning, match='The "MockSpider.h1" method is a generator'):
        warn_on_generator_with_return_value(mock_spider, h1)
    with pytest.warns(UserWarning, match='The "MockSpider.i1" method is a generator'):
        warn_on_generator_with_return_value(mock_spider, i1)


def test_generators_return_none(mock_spider):
    def f2():
        yield 1

    def g2():
        yield 1

    def h2():
        yield 1

    def i2():
        yield 1
        yield from generator_that_returns_stuff()

    def j2():
        yield 1

        def helper() -> int:
            return 0

        yield helper()

    def k2():
        """
        docstring
        """
        url = """
https://example.org
    """
        yield url

    def l2():
        return

    assert not is_generator_with_return_value(top_level_return_none)
    assert not is_generator_with_return_value(f2)
    assert not is_generator_with_return_value(g2)
    assert not is_generator_with_return_value(h2)
    assert not is_generator_with_return_value(i2)
    assert not is_generator_with_return_value(j2)  # not recursive
    assert not is_generator_with_return_value(k2)  # not recursive
    assert not is_generator_with_return_value(l2)

    with warnings.catch_warnings():
        warnings.simplefilter("error", UserWarning)
        warn_on_generator_with_return_value(mock_spider, top_level_return_none)
        warn_on_generator_with_return_value(mock_spider, f2)
        warn_on_generator_with_return_value(mock_spider, g2)
        warn_on_generator_with_return_value(mock_spider, h2)
        warn_on_generator_with_return_value(mock_spider, i2)
        warn_on_generator_with_return_value(mock_spider, j2)
        warn_on_generator_with_return_value(mock_spider, k2)
        warn_on_generator_with_return_value(mock_spider, l2)


def test_generators_return_none_with_decorator(mock_spider):
    def decorator(func):
        def inner_func():
            func()

        return inner_func

    @decorator
    def f3():
        yield 1

    @decorator
    def g3():
        yield 1

    @decorator
    def h3():
        yield 1

    @decorator
    def i3():
        yield 1
        yield from generator_that_returns_stuff()

    @decorator
    def j3():
        yield 1

        def helper() -> int:
            return 0

        yield helper()

    @decorator
    def k3():
        """
        docstring
        """
        url = """
https://example.org
    """
        yield url

    @decorator
    def l3():
        return

    assert not is_generator_with_return_value(top_level_return_none)
    assert not is_generator_with_return_value(f3)
    assert not is_generator_with_return_value(g3)
    assert not is_generator_with_return_value(h3)
    assert not is_generator_with_return_value(i3)
    assert not is_generator_with_return_value(j3)  # not recursive
    assert not is_generator_with_return_value(k3)  # not recursive
    assert not is_generator_with_return_value(l3)

    with warnings.catch_warnings():
        warnings.simplefilter("error", UserWarning)
        warn_on_generator_with_return_value(mock_spider, top_level_return_none)
        warn_on_generator_with_return_value(mock_spider, f3)
        warn_on_generator_with_return_value(mock_spider, g3)
        warn_on_generator_with_return_value(mock_spider, h3)
        warn_on_generator_with_return_value(mock_spider, i3)
        warn_on_generator_with_return_value(mock_spider, j3)
        warn_on_generator_with_return_value(mock_spider, k3)
        warn_on_generator_with_return_value(mock_spider, l3)


@mock.patch("scrapy.utils.misc.is_generator_with_return_value", new=_indentation_error)
def test_indentation_error(mock_spider):
    with pytest.warns(UserWarning, match="Unable to determine"):
        warn_on_generator_with_return_value(mock_spider, top_level_return_none)


def test_partial() -> None:
    def cb(arg1, arg2):
        yield {}

    partial_cb = partial(cb, arg1=42)
    assert not is_generator_with_return_value(partial_cb)


def test_warn_on_generator_with_return_value_settings_disabled() -> None:
    class MockSettings:
        def __init__(self, settings_dict: dict[str, Any] | None = None):
            self.settings_dict = settings_dict or {}

        def getbool(self, name, default=False):
            return self.settings_dict.get(name, default)

    class MockSpider:
        def __init__(self) -> None:
            self.settings = MockSettings({"WARN_ON_GENERATOR_RETURN_VALUE": False})

    spider = MockSpider()

    def gen_with_return():
        yield 1
        return "value"

    with warnings.catch_warnings():
        warnings.simplefilter("error", UserWarning)
        warn_on_generator_with_return_value(spider, gen_with_return)  # type: ignore[arg-type]

    spider.settings.settings_dict["WARN_ON_GENERATOR_RETURN_VALUE"] = True

    with pytest.warns(UserWarning, match="is a generator"):
        warn_on_generator_with_return_value(spider, gen_with_return)  # type: ignore[arg-type]



from __future__ import annotations

from typing import TYPE_CHECKING

from scrapy.settings import Settings
from scrapy.utils.job import job_dir

if TYPE_CHECKING:
    from pathlib import Path


def test_no_jobdir() -> None:
    assert job_dir(Settings()) is None
    assert job_dir(Settings({"JOBDIR": ""})) is None


def test_existing_jobdir(tmp_path: Path) -> None:
    assert job_dir(Settings({"JOBDIR": str(tmp_path)})) == str(tmp_path)


def test_missing_jobdir(tmp_path: Path) -> None:
    jobdir = tmp_path / "missing" / "jobdir"
    assert job_dir(Settings({"JOBDIR": str(jobdir)})) == str(jobdir)
    assert jobdir.is_dir()



"""
Scrapy core exceptions

These exceptions are documented in docs/topics/exceptions.rst. Please don't add
new exceptions here without documenting them there.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from scrapy.http import Response

# Internal


class NotConfigured(Exception):
    """Raised by a :ref:`component <topics-components>` from its ``__init__()``
    or :meth:`from_crawler` method to indicate that it will remain disabled.

    Only the following components can be disabled this way:

    -   :ref:`Downloader middlewares <topics-downloader-middleware>`
    -   :ref:`Extensions <topics-extensions>`
    -   :ref:`Item pipelines <topics-item-pipeline>`
    -   :ref:`Spider middlewares <topics-spider-middleware>`"""


class _InvalidOutput(TypeError):
    """
    Indicates an invalid value has been returned by a middleware's processing method.
    Internal and undocumented, it should not be raised or caught by user code.
    """


# HTTP and crawling


class IgnoreRequest(Exception):
    """Raised to indicate that a request should be ignored.

    A :ref:`downloader middleware <topics-downloader-middleware>` can raise it
    from its
    :meth:`~scrapy.downloadermiddlewares.DownloaderMiddleware.process_request`
    or
    :meth:`~scrapy.downloadermiddlewares.DownloaderMiddleware.process_response`
    method to drop a request, and a :signal:`request_scheduled` signal handler
    can raise it to drop a request before it reaches the
    :ref:`scheduler <topics-scheduler>`."""


class DontCloseSpider(Exception):
    """Raised in a :signal:`spider_idle` signal handler to prevent the spider
    from being closed."""


class CloseSpider(Exception):
    """Raised from a :ref:`spider callback <topics-spiders>`, or while the
    spider is starting, to request the spider to be closed/stopped.

    .. versionchanged:: 2.18.0
       Added support for raising it while the spider is starting.

    *reason* is a string with the reason for closing.

    For example:

    .. code-block:: python

        def parse_page(self, response):
            if "Bandwidth exceeded" in response.text:
                raise CloseSpider("bandwidth_exceeded")
    """

    def __init__(self, reason: str = "cancelled"):
        super().__init__()
        self.reason = reason


class StopDownload(Exception):
    """Raised from a :class:`~scrapy.signals.bytes_received` or
    :class:`~scrapy.signals.headers_received` signal handler to :ref:`stop the
    download <topics-stop-response-download>` of the response body.

    The ``fail`` boolean parameter controls which method will handle the
    resulting response:

    * If ``fail=True`` (default), the request errback is called. The response
      object is available as the ``response`` attribute of the ``StopDownload``
      exception, which is in turn stored as the ``value`` attribute of the
      received :class:`~twisted.python.failure.Failure` object. This means that
      in an errback defined as ``def errback(self, failure)``, the response can
      be accessed though ``failure.value.response``.

    * If ``fail=False``, the request callback is called instead.

    In both cases, the response could have its body truncated: the body contains
    all bytes received up until the exception is raised, including the bytes
    received in the signal handler that raises the exception. Also, the response
    object is marked with ``"download_stopped"`` in its
    :attr:`~scrapy.http.Response.flags` attribute.
    """

    response: Response | None

    def __init__(self, *, fail: bool = True):
        super().__init__()
        self.fail = fail


class DownloadConnectionRefusedError(Exception):
    """Indicates that a connection was refused by the server."""


class CannotResolveHostError(Exception):
    """Indicates that the provided hostname cannot be resolved."""


class DownloadTimeoutError(Exception):
    """Indicates that a request download has timed out."""


class DownloadCancelledError(Exception):
    """Indicates that a request download was cancelled."""


class DownloadFailedError(Exception):
    """Indicates that a request download has failed."""


class ResponseDataLossError(Exception):
    """Indicates that Scrapy couldn't get a complete response."""


class UnsupportedURLSchemeError(Exception):
    """Indicates that the URL scheme is not supported."""


# Items


class DropItem(Exception):
    """Raised from the :meth:`process_item` method of an :ref:`item pipeline
    <topics-item-pipeline>` to stop the processing of an item."""

    def __init__(self, message: str, log_level: str | None = None):
        super().__init__(message)
        self.log_level = log_level


class NotSupported(Exception):
    """Raised to indicate that a requested feature is not supported.

    For example, Scrapy raises it when text-parsing shortcuts such as
    :meth:`response.css() <scrapy.http.TextResponse.css>` or
    :meth:`response.xpath() <scrapy.http.TextResponse.xpath>` are used on a
    :class:`~scrapy.http.Response` whose content is not text, or when sending a
    request whose URL scheme has no matching :ref:`download handler
    <topics-download-handlers>`."""


# Commands


class UsageError(Exception):
    """To indicate a command-line usage error"""

    def __init__(self, *a: Any, **kw: Any):
        self.print_help = kw.pop("print_help", True)
        super().__init__(*a, **kw)


class ScrapyDeprecationWarning(Warning):
    """Warning category for deprecated features, since the default
    :exc:`DeprecationWarning` is silenced.
    """


class ContractFail(AssertionError):
    """Error raised in case of a failing contract"""



import xmlrpc.client
from typing import Any

import pytest

from scrapy import Request
from scrapy.http import XmlRpcRequest
from scrapy.utils.python import to_bytes
from tests.utils.bases.http_request import TestRequestBase


class TestRequest(TestRequestBase):
    request_class = Request


class TestXmlRpcRequest(TestRequestBase):
    request_class = XmlRpcRequest
    default_method = "POST"
    default_headers = {b"Content-Type": [b"text/xml"]}

    def _test_request(self, **kwargs: Any) -> None:
        r = self.request_class("http://scrapytest.org/rpc2", **kwargs)
        assert r.headers[b"Content-Type"] == b"text/xml"
        assert r.body == to_bytes(
            xmlrpc.client.dumps(**kwargs), encoding=kwargs.get("encoding", "utf-8")
        )
        assert r.method == "POST"
        assert r.encoding == kwargs.get("encoding", "utf-8")
        assert r.dont_filter

    def test_xmlrpc_dumps(self):
        self._test_request(params=("value",))
        self._test_request(params=("username", "password"), methodname="login")
        self._test_request(params=("response",), methodresponse="login")
        self._test_request(params=("pas£",), encoding="utf-8")
        self._test_request(params=(None,), allow_none=1)
        with pytest.raises(TypeError):
            self._test_request()
        with pytest.raises(TypeError):
            self._test_request(params=(None,))

    def test_latin1(self):
        self._test_request(params=("pas£",), encoding="latin1")



from __future__ import annotations

from typing import TYPE_CHECKING, cast

import pytest
from twisted.internet import defer

from scrapy import signals
from scrapy.core.engine import ExecutionEngine
from scrapy.statscollectors import MemoryStatsCollector
from scrapy.utils.spider import DefaultSpider
from scrapy.utils.test import get_crawler
from tests.utils.decorators import coroutine_test

if TYPE_CHECKING:
    from scrapy.core.scheduler import Scheduler
    from scrapy.crawler import Crawler


@pytest.fixture
def crawler() -> Crawler:
    crawler = get_crawler(DefaultSpider)
    crawler.spider = crawler._create_spider()
    return crawler


@coroutine_test
async def test_no_slot(crawler: Crawler) -> None:
    engine = ExecutionEngine(crawler, lambda _: None)
    crawler.engine = engine
    await engine.open_spider_async()
    slot = engine._slot
    engine._slot = None
    with pytest.raises(RuntimeError, match="Engine slot not assigned"):
        await engine.close_spider_async()
    # close it correctly
    engine._slot = slot
    await engine.close_spider_async()


@coroutine_test
async def test_no_spider(crawler: Crawler) -> None:
    engine = ExecutionEngine(crawler, lambda _: None)
    with pytest.raises(RuntimeError, match="Spider not opened"):
        await engine.close_spider_async()
    engine.downloader.close()  # cleanup


@coroutine_test
async def test_exception_slot(
    crawler: Crawler, caplog: pytest.LogCaptureFixture
) -> None:
    engine = ExecutionEngine(crawler, lambda _: None)
    crawler.engine = engine
    await engine.open_spider_async()
    assert engine._slot
    del engine._slot.heartbeat
    await engine.close_spider_async()
    assert "Slot close failure" in caplog.text


@coroutine_test
async def test_exception_downloader(
    crawler: Crawler, caplog: pytest.LogCaptureFixture
) -> None:
    engine = ExecutionEngine(crawler, lambda _: None)
    crawler.engine = engine
    await engine.open_spider_async()
    del engine.downloader.slots
    await engine.close_spider_async()
    assert "Downloader close failure" in caplog.text


@coroutine_test
async def test_exception_scraper(
    crawler: Crawler, caplog: pytest.LogCaptureFixture
) -> None:
    engine = ExecutionEngine(crawler, lambda _: None)
    crawler.engine = engine
    await engine.open_spider_async()
    engine.scraper.slot = None
    await engine.close_spider_async()
    assert "Scraper close failure" in caplog.text


@coroutine_test
async def test_exception_scheduler(
    crawler: Crawler, caplog: pytest.LogCaptureFixture
) -> None:
    engine = ExecutionEngine(crawler, lambda _: None)
    crawler.engine = engine
    await engine.open_spider_async()
    assert engine._slot
    del cast("Scheduler", engine._slot.scheduler).dqs
    await engine.close_spider_async()
    assert "Scheduler close failure" in caplog.text


@coroutine_test
async def test_exception_signal(
    crawler: Crawler, caplog: pytest.LogCaptureFixture
) -> None:
    engine = ExecutionEngine(crawler, lambda _: None)
    crawler.engine = engine
    await engine.open_spider_async()
    signal_manager = engine.signals
    del engine.signals
    await engine.close_spider_async()
    assert "Error while sending spider_close signal" in caplog.text
    # send the spider_closed signal to close various components
    await signal_manager.send_catch_log_async(
        signal=signals.spider_closed,
        spider=engine.spider,
        reason="cancelled",
    )


@coroutine_test
async def test_exception_stats(
    crawler: Crawler, caplog: pytest.LogCaptureFixture
) -> None:
    engine = ExecutionEngine(crawler, lambda _: None)
    crawler.engine = engine
    await engine.open_spider_async()
    assert isinstance(crawler.stats, MemoryStatsCollector)
    del crawler.stats.spider_stats
    await engine.close_spider_async()
    assert "Stats close failure" in caplog.text


@coroutine_test
async def test_exception_callback(
    crawler: Crawler, caplog: pytest.LogCaptureFixture
) -> None:
    engine = ExecutionEngine(crawler, lambda _: defer.fail(ValueError()))
    crawler.engine = engine
    await engine.open_spider_async()
    await engine.close_spider_async()
    assert "Error running spider_closed_callback" in caplog.text


@coroutine_test
async def test_exception_async_callback(
    crawler: Crawler, caplog: pytest.LogCaptureFixture
) -> None:
    async def cb(_):
        raise ValueError

    engine = ExecutionEngine(crawler, cb)
    crawler.engine = engine
    await engine.open_spider_async()
    await engine.close_spider_async()
    assert "Error running spider_closed_callback" in caplog.text



from __future__ import annotations

from shutil import copytree
from typing import TYPE_CHECKING

import pytest

from tests.utils.cmdline import call

if TYPE_CHECKING:
    from pathlib import Path


class TestProjectBase:
    """A base class for tests that may need a Scrapy project."""

    project_name = "testproject"

    @pytest.fixture(scope="session")
    def _proj_path_cached(self, tmp_path_factory: pytest.TempPathFactory) -> Path:
        """Create a Scrapy project in a temporary directory and return its path.

        Used as a cache for ``proj_path``.
        """
        tmp_path = tmp_path_factory.mktemp("proj")
        call("startproject", self.project_name, cwd=tmp_path)
        return tmp_path / self.project_name

    @pytest.fixture
    def proj_path(self, tmp_path: Path, _proj_path_cached: Path) -> Path:
        """Copy a pre-generated Scrapy project into a temporary directory and return its path."""
        proj_path = tmp_path / self.project_name
        copytree(_proj_path_cached, proj_path)
        return proj_path

    @staticmethod
    def _append_settings(proj_mod_path: Path, text: str) -> None:
        """Add text to the end of the project settings.py."""
        with (proj_mod_path / "settings.py").open("a", encoding="utf-8") as f:
            f.write(text)

    @classmethod
    def _break_bootstrap(cls, proj_mod_path: Path) -> None:
        """Enable a spider middleware that raises an exception when built, so
        that crawling any spider of the project fails before it starts."""
        cls._append_settings(
            proj_mod_path,
            """

class BreakingMiddleware:
    def __init__(self):
        raise RuntimeError("Breaking middleware")


SPIDER_MIDDLEWARES = {f"{__name__}.BreakingMiddleware": 100}
""",
        )
